"""Handler for selecting and running editor features"""

from typing import Any, Union

from . import (
    helper,
    user_input_handler,
    config_manager,
)
from .edits import basic, cats, gamototo, levels, other, save_management


def fix_elsewhere_old(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Fix the elsewhere error using 2 save files"""

    main_token = save_stats["token"]
    main_iq = save_stats["inquiry_code"]
    input(
        "Select a save file that is currently loaded in-game that doesn't have the elsehere error and is not banned\nPress enter to continue:"
    )
    new_path = helper.select_file(
        "Select a clean save file",
        helper.get_save_file_filetype(),
        helper.get_save_path(),
    )
    if not new_path:
        print("Please select a save file")
        return save_stats

    data = helper.load_save_file(new_path)
    new_stats = data["save_stats"]
    new_token = new_stats["token"]
    new_iq = new_stats["inquiry_code"]
    save_stats["token"] = new_token
    save_stats["inquiry_code"] = new_iq

    helper.colored_text(f"Replaced inquiry code: &{main_iq}& with &{new_iq}&")
    helper.colored_text(f"Replaced token: &{main_token}& with &{new_token}&")
    return save_stats


FEATURES: dict[str, Any] = {
    "Save Management": {
        "Save changes and upload to game servers (get transfer and confirmation codes)": save_management.server_upload.save_and_upload,
        "Save changes and exit": save_management.save_and_exit.save_and_exit,
        "Save changes and push save data to the game (don't re-open game)": save_management.save_and_exit.save_and_push,
        "Save changes and push save data to the game (re-open game)": save_management.save_and_exit.save_and_push_rerun,
        "Export save data as json": save_management.other.export,
        "Clear save data (used to generate a new account without re-installing the game)": save_management.other.clear_data,
        "Upload tracked bannable items (This is done automatically when saving and exiting)": save_management.server_upload.upload_metadata,
    },
    "Items": {
        "Cat Food": basic.basic_items.edit_cat_food,
        "XP": basic.basic_items.edit_xp,
        "Tickets": {
            "Normal Tickets": basic.basic_items.edit_normal_tickets,
            "Rare Tickets": basic.basic_items.edit_rare_tickets,
            "Platinum Tickets": basic.basic_items.edit_platinum_tickets,
            "Platinum Shards": basic.basic_items.edit_platinum_shards,
            "Legend Tickets": basic.basic_items.edit_legend_tickets,
        },
        "NP": basic.basic_items.edit_np,
        "Leadership": basic.basic_items.edit_leadership,
        "Battle Items": basic.basic_items.edit_battle_items,
        "Catseyes": basic.basic_items.edit_catseyes,
        "Cat Fruit / Behemoth Stones": basic.catfruit.edit_catfruit,
        "Talent Orbs": basic.talent_orbs.edit_talent_orbs,
        "Catamins": basic.basic_items.edit_catamins,
    },
    "Gamatoto / Ototo": {
        "Ototo Engineers": basic.basic_items.edit_engineers,
        "Base materials": basic.basic_items.edit_base_materials,
        "Catamins": basic.basic_items.edit_catamins,
        "Gamatoto XP / Level": gamototo.gamatoto_xp.edit_gamatoto_xp,
        "Ototo Cat Cannon": gamototo.ototo_cat_cannon.edit_cat_cannon,
        "Gamatoto Helpers": gamototo.helpers.edit_helpers,
        "Fix gamatoto from crashing the game": gamototo.fix_gamatoto.fix_gamatoto,
    },
    "Cats / Special Skills": {
        "Get / Remve Cats": {
            "Get Cats": cats.get_remove_cats.get_cat,
            "Remove Cats": cats.get_remove_cats.remove_cats,
        },
        "Upgrade Cats": cats.upgrade_cats.upgrade_cats,
        "True Form Cats": {
            "Get Cat True Forms": cats.evolve_cats.get_evolve,
            "Remove Cat True Forms": cats.evolve_cats.remove_evolve,
            "Force True Form Cats (will lead to blank cats for cats without a true form)": cats.evolve_cats.get_evolve_forced,
        },
        "Talents": {
            "Set talents for each selected cat individually": cats.talents.edit_talents_individual,
            "Max / Remove all selected cat talents": cats.talents.max_all_talents,
        },
        "Collect / Remove Cat Guide": {
            "Set Cat Guide Entries (does not give cf)": cats.clear_cat_guide.collect_cat_guide,
            "Unclaim Cat Guide Entries": cats.clear_cat_guide.remove_cat_guide,
        },
        'Get stage unit drops - removes the "Clear this stage to get special cat" dialog': cats.chara_drop.get_character_drops,
        "Upgrade special skills / abilities": cats.upgrade_blue.upgrade_blue,
    },
    "Levels / Treasures": {
        "Main Story Chapters Clear / Unclear": {
            "Clear each stage in every chapter for all selected chapters": levels.main_story.clear_all,
            "Clear each stage in every chapter for each selected chapter": levels.main_story.clear_each,
        },
        "Treasures": {
            "Treasure Groups (e.g energy drink, aqua crystal, etc)": levels.treasures.treasure_groups,
            "Specific stages and specific chapters individually": levels.treasures.specific_stages,
            "Specific stages and chapters all at once": levels.treasures.specific_stages_all_chapters,
        },
        "Zombie Stages / Outbreaks": levels.outbreaks.edit_outbreaks,
        "Event Stages": levels.event_stages.event_stages,
        "Stories of Legend": levels.event_stages.stories_of_legend,
        "Uncanny Legends": levels.uncanny.edit_uncanny,
        "Aku Realm/Gates": levels.aku.edit_aku,
        "Gauntlets": levels.gauntlet.edit_gauntlet,
        "Into the Future Timed Scores": levels.itf_timed_scores.timed_scores,
        "Challenge Battle Score": basic.basic_items.edit_challenge_battle,
        "Towers": levels.towers.edit_tower,
        "Clear Tutorial": levels.clear_tutorial.clear_tutorial,
        "Catclaw Dojo Score (Hall of Initiates)": basic.basic_items.edit_dojo_score,
    },
    "Inquiry Code / Token / Account": {
        "Inquiry Code": basic.basic_items.edit_inquiry_code,
        "Token": basic.basic_items.edit_token,
        "Fix elsewhere error / Unban account": other.fix_elsewhere.fix_elsewhere,
        "Old Fix elsewhere error / Unban account (needs 2 save files)": fix_elsewhere_old,
        "Create a new account / Duplicate account": other.create_new_account.create_new_account,
    },
    "Other": {
        "Rare Gacha Seed": basic.basic_items.edit_rare_gacha_seed,
        "Unlocked Equip Slots": basic.basic_items.edit_unlocked_slots,
        "Get Restart Pack / Returner Mode": basic.basic_items.edit_restart_pack,
        "Meow Medals": other.meow_medals.medals,
        "Play Time": other.play_time.edit_play_time,
        "Unlock / Remove Enemy Guide Entries": other.unlock_enemy_guide.enemy_guide,
        "Catnip Challenges / Missions": other.missions.edit_missions,
        "Normal Ticket Max Trade Progress (allows for unbannable rare tickets)": other.trade_progress.set_trade_progress,
        "Unlock the Equip Menu": other.unlock_equip_menu.unlock_equip,
        "Get Gold Pass": other.get_gold_pass.get_gold_pass,
        "Claim / Remove all user rank rewards (does not give any items)": other.claim_user_rank_rewards.edit_rewards,
    },
    "Edit Config" : {
        "Edit DEFAULT_GAME_VERSION" : config_manager.edit_default_gv,
        "Edit DEFAULT_SAVE_PATH" : config_manager.edit_default_save_file_path,
        "Edit EDITOR settings" : config_manager.edit_editor_settings,
        "Edit START_UP settings" : config_manager.edit_start_up_settings,
        "Edit SAVE_CHANGES settings" : config_manager.edit_save_changes_settings,
        "Edit SERVER settings" : config_manager.edit_server_settings,
    }
} 


def get_feature(
    selected_features: Any, search_string: str, results: dict[str, Any]
) -> dict[str, Any]:
    """Search for a feature if the feature name contains the search string"""

    for feature in selected_features:
        feature_data = selected_features[feature]
        if isinstance(feature_data, dict):
            feature_data = get_feature(feature_data, search_string, results)
        if search_string.lower().replace(" ", "") in feature.lower().replace(" ", ""):
            results[feature] = selected_features[feature]
    return results


def show_options(
    save_stats: dict[str, Any], features_to_use: dict[str, Any]
) -> dict[str, Any]:
    """Allow the user to either enter a feature number or a feature name, and get the features that match"""

    if (
        not config_manager.get_config_value_category("EDITOR", "SHOW_CATEGORIES")
        and FEATURES == features_to_use
    ):
        user_input = ""
    else:
        prompt = (
            "What do you want to edit (some options contain other features within them)"
        )
        if config_manager.get_config_value_category(
            "EDITOR", "SHOW_FEATURE_SELECT_EXPLANATION"
        ):
            prompt += "\nYou can enter a number to run a feature or a word to search for that feature (e.g entering catfood will run the Cat Food feature, and entering tickets will show you all the features that edit tickets)\nYou can press enter to see a list of all of the features"
        user_input = user_input_handler.colored_input(f"{prompt}:\n")
    user_int = helper.check_int(user_input)
    results = []
    if user_int is None:
        results = get_feature(features_to_use, user_input, {})
    else:
        if user_int < 1 or user_int > len(features_to_use) + 1:
            helper.colored_text("Value out of range", helper.RED)
            return show_options(save_stats, features_to_use)
        if FEATURES != features_to_use:
            if user_int - 2 < 0:
                return menu(save_stats)
            results = features_to_use[list(features_to_use)[user_int - 2]]
        else:
            results = features_to_use[list(features_to_use)[user_int - 1]]
    if not isinstance(results, dict):
        save_stats_return = results(save_stats)
        if save_stats_return is None:
            return save_stats
        return save_stats_return
    if len(results) == 0:
        helper.colored_text("No feature found with that name.", helper.RED)
        return menu(save_stats)
    if len(results) == 1 and isinstance(list(results.values())[0], dict):
        results = results[list(results)[0]]
    if len(results) == 1:
        save_stats_return = results[list(results)[0]](save_stats)
        if save_stats_return is None:
            return save_stats
        return save_stats_return

    helper.colored_list(["Go Back"] + list(results))
    return show_options(save_stats, results)


def menu(
    save_stats: dict[str, Any], path_save: Union[str, None] = None
) -> dict[str, Any]:
    """Show the menu and allow the user to select a feature to edit"""

    if path_save:
        helper.set_save_path(path_save)
    if config_manager.get_config_value_category("EDITOR", "SHOW_CATEGORIES"):
        helper.colored_list(list(FEATURES))
    save_stats = show_options(save_stats, FEATURES)

    return save_stats



"""Handler for clearing gauntlets"""
from typing import Any

from . import event_stages
from ... import user_input_handler, helper
from ..other import meow_medals

def edit_gauntlet(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for clearing gauntlets"""

    stage_data = save_stats["gauntlets"]
    lengths = stage_data["Lengths"]

    ids = []
    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter gauntlet ids (Look up &Event Release Order battle cats& and scroll past the &events& to find &gauntlet& ids) (You can enter &all& to get all, a range e.g 1-49, or ids separate by spaces e.g &5 4 7&):"
        ),
        lengths["total"],
    )
    save_stats["gauntlets"] = event_stages.stage_handler(stage_data, ids, 0)
    base_addr = meow_medals.BaseMapIds.GAUNTLETS.value
    (
        save_stats["gauntlets"],
        save_stats["medals"],
    ) = event_stages.set_medals(
        save_stats["gauntlets"],
        save_stats["medals"],
        (base_addr, base_addr + len(save_stats["gauntlets"]["Value"]["unlock_next"])),
        -base_addr,
        helper.check_data_is_jp(save_stats),
    )
    return save_stats



"""Handler for clearing event stages"""
from typing import Any

from ... import user_input_handler, helper
from ...edits.other import meow_medals

def set_stage_data(
    stage_data_edit: dict[str, Any],
    stage_id: int,
    stars: int,
    lengths: dict[str, int],
    unlock_next: bool,
) -> dict[str, Any]:
    """Set the stage data for a stage"""

    if stage_id >= len(stage_data_edit["Value"]["clear_progress"]):
        return stage_data_edit
    stage_data_edit = set_clear_progress(stage_data_edit, stage_id, stars, lengths)
    if unlock_next and stage_id + 1 < len(stage_data_edit["Value"]["clear_progress"]):
        stage_data_edit = set_unlock_next(stage_data_edit, stage_id, stars, lengths)
    stage_data_edit = set_clear_amount(stage_data_edit, stage_id, stars, lengths)
    return stage_data_edit


def set_clear_progress(
    stage_data: dict[str, Any], stage_id: int, stars: int, lengths: dict[str, int]
) -> dict[str, Any]:
    """Set the clear progress for a stage"""

    stage_data["Value"]["clear_progress"][stage_id] = ([lengths["stages"]] * stars) + (
        [0] * (lengths["stars"] - stars)
    )
    return stage_data


def set_unlock_next(
    stage_data: dict[str, Any], stage_id: int, stars: int, lengths: dict[str, int]
) -> dict[str, Any]:
    """Set the unlock next for a stage"""

    stage_data["Value"]["unlock_next"][stage_id + 1] = (
        [lengths["stars"] - 1] * stars
    ) + ([0] * (lengths["stars"] - stars))
    return stage_data


def set_clear_amount(
    stage_data: dict[str, Any], stage_id: int, stars: int, lengths: dict[str, int]
) -> dict[str, Any]:
    """Set the clear amount for a stage"""

    stage_data["Value"]["clear_amount"][stage_id] = (
        [[1] * lengths["stages"]] * stars
    ) + ([[0] * lengths["stages"]] * (lengths["stars"] - stars))
    return stage_data


def set_medals(
    stage_stats: dict[str, Any],
    medal_stats: dict[str, Any],
    valid_range: tuple[int, int],
    offset: int,
    is_jp: bool,
) -> tuple[dict[str, Any], dict[str, Any]]:
    """Set the medals for completed stages"""

    medal_data = meow_medals.get_medal_data(is_jp)

    unlock_next = stage_stats["Value"]["unlock_next"]

    for medal in medal_data.stages:
        if not medal.maps:
            continue
        completed = True
        for map_id in medal.maps:
            star = medal.star
            if map_id < 0:
                continue
            if map_id < valid_range[0] or map_id > valid_range[1]:
                completed = False
                break
            map_id += offset
            next_chapter = unlock_next[map_id + 1]
            if star is None:
                star = 0
            if next_chapter[star] == 0:
                completed = False
                break
        if completed:
            if medal.medal_id not in medal_stats["medal_data_1"]:
                medal_stats["medal_data_1"].append(medal.medal_id)
            medal_stats["medal_data_2"][medal.medal_id] = 1
    return stage_stats, medal_stats


def stage_handler(
    stage_data: dict[str, Any], ids: list[int], offset: int, unlock_next: bool = True
) -> dict[str, Any]:
    """Clear stages from a set of ids"""

    lengths = stage_data["Lengths"]

    individual = True
    if len(ids) > 1:
        individual = user_input_handler.ask_if_individual("stars / crowns for each stage")
    first = True
    stars = 0
    stage_data_edit = stage_data
    for stage_id in ids:
        if not individual and first:
            stars = helper.check_int(
                user_input_handler.colored_input(
                    f"Enter the number of stars/cowns (max &{lengths['stars']}&):"
                )
            )
            if stars is None:
                print("Please enter a valid number")
                break
            stars = helper.clamp(stars, 0, lengths["stars"])
            first = False
        elif individual:
            stars = helper.check_int(
                user_input_handler.colored_input(
                    f"Enter the number of stars/cowns for subchapter &{stage_id}& (max &{lengths['stars']}&):"
                )
            )
            if stars is None:
                print("Please enter a valid number")
                break
            stars = helper.clamp(stars, 0, lengths["stars"])
        stage_id += offset
        stage_data_edit = stage_data
        stage_data_edit = set_stage_data(
            stage_data_edit, stage_id, stars, lengths, unlock_next
        )

    print("Successfully set subchapters")

    return stage_data_edit


def stories_of_legend(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for clearing stories of legend"""

    stage_data = save_stats["event_stages"]

    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter subchapter ids (e.g &1& = legend begins, &2& = passion land)(You can enter &all& to get all, a range e.g &1&-&49&, or ids separate by spaces e.g &5 4 7&):"
        ),
        50,
    )
    offset = -1
    save_stats["event_stages"] = stage_handler(stage_data, ids, offset)
    save_stats["event_stages"], save_stats["medals"] = set_medals(save_stats["event_stages"], save_stats["medals"], (0, 50), 0, helper.check_data_is_jp(save_stats))
    return save_stats


def event_stages(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for clearing event stages"""

    stage_data = save_stats["event_stages"]
    lengths = stage_data["Lengths"]

    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter subchapter ids (Look up &Event Release Order battle cats& to find ids)(You can enter &all& to get all, a range e.g &1&-&50&, or ids separate by spaces e.g &5 4 7&):"
        ),
        lengths["total"] - 400,
    )
    offset = 400
    save_stats["event_stages"] = stage_handler(stage_data, ids, offset)
    save_stats["event_stages"], save_stats["medals"] = set_medals(save_stats["event_stages"], save_stats["medals"], (0, len(save_stats["event_stages"]["Value"]["unlock_next"])), -600, helper.check_data_is_jp(save_stats))
    return save_stats



"""Item object"""

from typing import Any, Union

from . import helper, user_input_handler, managed_item, config_manager, tracker


class BannableItem:
    """BannableItem object"""

    def __init__(
        self,
        is_bannable: bool = False,
        has_workaround: bool = False,
        work_around_text: str = "",
        managed_item_type: Union[None, managed_item.ManagedItemType] = None,
    ) -> None:
        self.is_bannable = is_bannable
        self.has_workaround = has_workaround
        self.work_around_text = work_around_text
        self.managed_item_type = managed_item_type


class Item:
    """An item"""

    def __init__(
        self,
        name: str,
        value: Union[int, str],
        max_value: Union[int, None],
        edit_name: str,
        bannable: BannableItem = BannableItem(False, False, ""),
        offset: int = 0,
        length: int = 4,
        set_name: str = "set",
        success_message: str = "",
        to: str = "to",
    ) -> None:
        self.name = name
        self.value = value
        self.max_value = max_value
        self.offset = offset
        self.edit_name = edit_name
        self.bannable = bannable
        self.length = length
        self.set_name = set_name
        if success_message:
            self.success_message = success_message
        else:
            self.success_message = set_name
        self.to = to

        if config_manager.get_config_value_category("EDITOR", "DISABLE_MAXES"):
            self.max_value = None

    def clamp(self) -> None:
        """Clamp the value to the max value"""

        if self.max_value is not None:
            self.value = helper.clamp(
                value=int(self.value),
                min_value=0,
                max_value=self.max_value,
            )
        else:
            # Make sure the value is not negative and fits in the byte length
            self.value = helper.clamp(
                value=int(self.value),
                min_value=0,
                max_value=256**self.length - 1,
            )

    def ban_warning(self) -> bool:
        """Warning for editing a bannable item"""

        if not config_manager.get_config_value_category("EDITOR", "SHOW_BAN_WARNING"):
            return False

        helper.colored_text(
            text=f"WARNING: Editing in {self.name} will most likely lead to a ban!",
            base=helper.RED,
        )
        if self.bannable.has_workaround:
            helper.colored_text(
                text=self.bannable.work_around_text,
                new=helper.RED,
            )
        leave = (
            user_input_handler.colored_input("Do you want to continue? (&y&/&n&):")
            == "n"
        )
        return leave

    def edit_int(self) -> None:
        """Handler for editing an integer value"""

        if self.bannable.is_bannable:
            leave = self.ban_warning()
            if leave:
                return
        if self.value is not None:
            helper.colored_text(
                f"The current {self.edit_name} of {self.name} is : &{int(self.value)+self.offset}&"
            )
        if self.max_value is None:
            max_str = "(max &None&)"
        else:
            max_str = f"(max &{self.max_value+self.offset}&)"

        text = f"Enter the {self.edit_name} of {self.name} you want to {self.set_name} {max_str}:"
        self.value = user_input_handler.get_int(text) - self.offset
        self.clamp()
        helper.colored_text(
            f"Successfully {self.success_message} the {self.edit_name} of {self.name} {self.to} &{self.value+self.offset}&"
        )

    def edit_str(self) -> None:
        """Handler for editing a string value"""

        helper.colored_text(
            f"The current {self.edit_name} of {self.name} is : &{self.value}&"
        )
        text = f"Enter the {self.edit_name} of {self.name} you want to {self.set_name}:"
        self.value = user_input_handler.colored_input(text)
        helper.colored_text(
            f"Successfully {self.success_message} the {self.edit_name} of {self.name} {self.to} &{self.value}&"
        )

    def edit(self) -> None:
        """Handler for editing an item"""

        original_val = self.value

        if isinstance(self.value, int):
            self.edit_int()
        elif self.value is None:
            self.edit_int()
        else:
            self.edit_str()
            return

        if self.bannable.is_bannable and original_val != self.value:
            new_val = self.value
            if not self.bannable.managed_item_type or isinstance(original_val, str):
                return
            item_tracker = tracker.Tracker()
            item_tracker.update_tracker(
                new_val - original_val, self.bannable.managed_item_type
            )


class ItemGroup:
    """A group of items"""

    def __init__(self, name: str, items: list[Item]):
        self.name = name
        self.items = items
        self.selected_items: list[Item] = []

    @property
    def max_value(self) -> Union[int, None]:
        """Get the maximum value of the selected items"""
        if config_manager.get_config_value_category("EDITOR", "DISABLE_MAXES"):
            return None
        try:
            return max(
                [
                    item.max_value
                    for item in self.selected_items
                    if item.max_value is not None
                ]
            )
        except (ValueError, TypeError):
            return None

    @property
    def selected_names(self) -> list[str]:
        """Get the names of the selected items"""
        return [item.name for item in self.selected_items]

    @property
    def names(self) -> list[str]:
        """Get the names of the items"""
        return [item.name for item in self.items]

    @property
    def selected_values(self) -> list[Union[int, str]]:
        """Get the values of the selected items"""
        return [item.value for item in self.selected_items]

    @property
    def values(self) -> list[Any]:
        """Get the values of the items"""
        return [item.value for item in self.items]

    def select(self) -> bool:
        """Select items from the group"""

        helper.colored_text(f"What &{self.name}& do you want to set?")
        ids = user_input_handler.select_inc(
            self.names,
            offset=self.items[0].offset,
            extra_data=self.values,
        )
        for option_id in ids[0]:
            if option_id < 0:
                helper.colored_text(
                    "You can't select options less than 1!",
                    base=helper.RED,
                )
                continue
            if option_id >= len(self.items):
                helper.colored_text(
                    "You can't select options greater than the last option!",
                    base=helper.RED,
                )
                continue
            self.selected_items.append(self.items[option_id])
        return ids[1]

    def edit(self) -> None:
        """Edit items from the group"""

        individual = self.select()
        if not individual:
            self.edit_all()
        else:
            for item in self.selected_items:
                print()
                item.edit()

    def edit_all(self) -> None:
        """Handler for editing all selected items at once"""

        if self.max_value is None:
            max_str = ""
        else:
            max_str = f" (max &{self.max_value+self.selected_items[0].offset}&)"
        text = f"Enter the {self.selected_items[0].edit_name} of {self.name} you want to set{max_str}:"
        value = user_input_handler.get_int(text) - self.selected_items[0].offset
        display_val = value
        for item in self.selected_items:
            item.value = value
            if self.max_value is not None:
                item.clamp()
                display_val = item.value

        helper.colored_text(
            f"Successfully set the {self.selected_items[0].success_message} of {self.name} to &{display_val + self.selected_items[0].offset}&"
        )


def create_item_group(
    names: list[str],
    values: Any,
    maxes: Any,
    edit_name: str,
    group_name: str,
    offset: int = 0,
) -> ItemGroup:
    """Create a list of items"""

    if isinstance(maxes, int) or maxes is None:
        maxes = [maxes] * len(names)
    if values is None:
        values = [None] * len(names)

    items = [
        Item(
            name=name,
            value=value,
            max_value=max_value,
            edit_name=edit_name,
            offset=offset,
        )
        for name, value, max_value in zip(names, values, maxes)
    ]

    item_group = ItemGroup(name=group_name, items=items)

    return item_group



from typing import Any

from ... import item, csv_handler, game_data_getter, helper


def get_fruit_names(is_jp: bool) -> list[str]:
    """Get the catfruit fruit names"""

    item_names = csv_handler.parse_csv(
        game_data_getter.get_file_latest("resLocal", "GatyaitemName.csv", is_jp).decode(
            "utf-8"
        ),
        delimeter=helper.get_text_splitter(is_jp),
    )
    fruit_ids = helper.parse_int_list_list(
        csv_handler.parse_csv(
            game_data_getter.get_file_latest("DataLocal", "Matatabi.tsv", is_jp).decode(
                "utf-8"
            ),
            delimeter="\t",
        )
    )[1:]
    fruit_names: list[str] = []
    for fruit in fruit_ids:
        fruit_names.append(item_names[fruit[0]][0])
    return fruit_names


def edit_catfruit(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing catruit"""

    max_cf = 128
    if save_stats["game_version"]["Value"] >= 110400:
        max_cf = None

    catfruit = item.create_item_group(
        names=get_fruit_names(helper.check_data_is_jp(save_stats)),
        values=save_stats["cat_fruit"],
        maxes=max_cf,
        edit_name="value",
        group_name="Catfruit",
    )
    catfruit.edit()
    save_stats["cat_fruit"] = catfruit.values
    return save_stats



"""Module that runs when the module is run directly"""

import os
import sys

from . import (
    adb_handler,
    config_manager,
    feature_handler,
    game_data_getter,
    helper,
    parse_save,
    patcher,
    serialise_save,
    server_handler,
    tracker,
    updater,
)
from .edits.levels import clear_tutorial


def print_start_up():
    """Print start up message"""

    helper.colored_text(
        "Welcome to the &Battle Cats Save File Editor&\n"
        + "Made by &fieryhenry&\n\n"
        + "GitHub: &https://github.com/fieryhenry/BCSFE-Python&\n"
        + "Discord: &https://discord.gg/DvmMgvn5ZB& - Please report any bugs to &#bug-reports&, or any suggestions to &#suggestions&\n"
        + "Donate: &https://ko-fi.com/fieryhenry&\n"
        + f"Config file path: &{os.path.join(config_manager.get_app_data_folder(), 'config.yaml')}&",
        base=helper.CYAN,
        new=helper.WHITE,
    )
    local_version = updater.get_local_version()

    print()
    if "b" in local_version:
        helper.colored_text(
            "You are using a &beta& release, some things may be broken. Please report any bugs you find to &#bug-reports& on Discord and specify that you are using a beta version",
            base=helper.RED,
            new=helper.WHITE,
        )
    print()
    helper.colored_text(
        "Thanks To:\n"
        + "&Lethal's editor& for giving me inspiration to start the project and it helped me work out how to patch the save data and edit cf/xp: &https://www.reddit.com/r/BattleCatsCheats/comments/djehhn/editoren&\n"
        + "&Beeven& and &csehydrogen's& code, which helped me figure out how to patch save data: &https://github.com/beeven/battlecats& and &https://github.com/csehydrogen/BattleCatsHacker&\n"
        + "Anyone who has supported my work for giving me motivation to keep working on this project: &https://ko-fi.com/fieryhenry&\n"
        + "Everyone in the discord for giving me saves, reporting bugs, suggesting new features, and for being an amazing community: &https://discord.gg/DvmMgvn5ZB&",
        base=helper.GREEN,
        new=helper.WHITE,
    )


def check_update() -> None:
    """Check if there is an update available and if so, ask the user if they want to update"""
    version_info = updater.get_version_info()
    stable_ver, pre_release_ver = version_info

    local_version = updater.get_local_version()

    helper.colored_text(
        f"Local version: &{local_version} | &Latest stable version: &{stable_ver}",
        base=helper.CYAN,
        new=helper.WHITE,
        end="",
    )
    if pre_release_ver > stable_ver:
        helper.colored_text(
            f"& | &Latest pre-release version: &{pre_release_ver}&",
            base=helper.CYAN,
            new=helper.WHITE,
            end="",
        )
    print()
    update_data = updater.check_update(version_info)
    if update_data[0]:
        helper.colored_text(
            "\nAn update is available, would you like to update? (&y&/&n&):",
            base=helper.GREEN,
            new=helper.WHITE,
            end="",
        )
        if input().lower() == "y":
            updater.update(update_data[1])
            helper.colored_text("Update successful", base=helper.GREEN)
            helper.exit_editor()


def main():
    """Main function"""

    item_tracker = tracker.Tracker()

    if config_manager.get_config_value_category(
        "SERVER", "WIPE_TRACKED_ITEMS_ON_START"
    ):
        item_tracker.reset_tracker()
    game_data_getter.check_remove_handler()

    check_updates = config_manager.get_config_value_category(
        "START_UP", "CHECK_FOR_UPDATES"
    )
    show_start = not config_manager.get_config_value_category(
        "START_UP", "HIDE_START_TEXT"
    )

    if show_start or check_updates:
        print()
        helper.print_line_seperator(helper.CYAN, length=200)
    if check_updates:
        check_update()
    if show_start:
        print_start_up()
    if show_start or check_updates:
        helper.print_line_seperator(helper.CYAN, length=200)
    normal_start_up()


def normal_start_up(default_op: bool = True) -> None:
    """Display and handle options for downloading save data, pulling save data, selecting save data"""

    default_start_option = config_manager.get_config_value_category(
        "START_UP", "DEFAULT_START_OPTION"
    )

    if default_start_option != -1 and default_op:
        option = str(default_start_option)
    else:
        print()
        if not default_op:
            helper.print_line_seperator(helper.WHITE)
        options = [
            "Download save data from the game using transfer and confirmation codes",
            "Select a save file from file",
            "Use adb to pull the save from a rooted device",
            "Load save data from json",
        ]
        helper.colored_list(options)
        option = input(f"Enter an option (1 to {len(options)}):")
    path = None
    if option == "1":
        print("Enter details for data transfer:")
        path = server_handler.download_handler()
    elif option == "2":
        print("Select save file:")
        path = helper.select_file(
            "Select a save file:",
            helper.get_save_file_filetype(),
            initial_file=helper.get_default_save_name(),
        )
    elif option == "3":
        print("Enter details for save pulling:")
        game_version = helper.ask_cc()
        path = adb_handler.adb_pull_save_data(game_version)
    elif option == "4":
        print("Select save data json file")
        js_path = helper.select_file(
            "Select save data json file",
            [("Json", "*.json")],
            initial_file=helper.get_default_save_name() + ".json",
        )
        if js_path:
            path = helper.load_json_handler(js_path)
    else:
        helper.colored_text("Please enter a recognised option", base=helper.RED)
        return normal_start_up(False)
    if not path:
        return normal_start_up(False)
    start(path)
    return None


def start(path: str) -> None:
    """Parse, patch, start the editor and serialise the save data"""
    data = helper.load_save_file(path)
    save_stats = data["save_stats"]
    save_data: bytes = data["save_data"]
    country_code = data["country_code"]

    if path.endswith(".json"):
        input(
            "Your save data seems to be in json format. Please use to import json option if you want to load json data.\nPress enter to continue...:"
        )
    if not clear_tutorial.is_tutorial_cleared(save_stats):
        save_stats = clear_tutorial.clear_tutorial(save_stats)
        save_data = serialise_save.start_serialize(save_stats)
    while True:
        save_stats = parse_save.start_parse(save_data, country_code)
        save_data = patcher.patch_save_data(save_data, country_code)
        save_stats = feature_handler.menu(save_stats, path)
        save_data = serialise_save.start_serialize(save_stats)
        save_data = patcher.patch_save_data(save_data, country_code)
        if config_manager.get_config_value_category(
            "SAVE_CHANGES", "SAVE_CHANGES_ON_EDIT"
        ):
            helper.write_file_bytes(path, save_data)
            helper.colored_text(
                f"Save data saved to &{path}&", base=helper.GREEN, new=helper.WHITE
            )
        if config_manager.get_config_value_category(
            "SAVE_CHANGES", "ALWAYS_EXPORT_JSON"
        ):
            helper.export_json(save_stats, path + ".json")


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit()



"""Handler for saving and exiting the editor"""

from typing import Any

from ... import helper, serialise_save, patcher, adb_handler


def save_and_exit(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Serialise the save data and exit"""

    save_data = serialise_save.start_serialize(save_stats)
    helper.write_save_data(
        save_data, save_stats["version"], helper.get_save_path(), True
    )

    helper.check_tracker(save_stats, helper.get_save_path())
    helper.ask_exit_editor()

    return save_stats


def save_and_push(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Serialise the save data and and push it to the game"""

    save_data = serialise_save.start_serialize(save_stats)
    save_data = patcher.patch_save_data(save_data, save_stats["version"])
    helper.write_file_bytes(helper.get_save_path(), save_data)

    helper.check_tracker(save_stats, helper.get_save_path())

    adb_handler.adb_push_save_data(save_stats["version"], helper.get_save_path())

    helper.ask_exit_editor()

    return save_stats


def save_and_push_rerun(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Serialise the save data and push it to the game and restart the game"""

    save_data = serialise_save.start_serialize(save_stats)
    save_data = patcher.patch_save_data(save_data, save_stats["version"])
    helper.write_file_bytes(helper.get_save_path(), save_data)

    helper.check_tracker(save_stats, helper.get_save_path())

    adb_handler.adb_push_save_data(save_stats["version"], helper.get_save_path())

    adb_handler.rerun_game(save_stats["version"])

    helper.ask_exit_editor()

    return save_stats



"""Handler for parsing CSV files."""


from typing import Any


def remove_pkcs7_padding(data: bytes) -> bytes:
    """Remove pkcs7 padding from data."""

    if len(data) % 16 != 0:
        raise Exception("Invalid data length")

    padding_length = data[-1]
    if padding_length > 16:
        raise Exception("Invalid padding length")
    if data[-padding_length:] != bytes([padding_length] * padding_length):
        raise Exception("Invalid padding")

    return data[:-padding_length]


def remove_comments(data: str) -> str:
    """Remove in-line comments from data."""

    data_ls = data.split("\n")
    data_ls = [line.split("//")[0] for line in data_ls]
    data_ls = [line.strip() for line in data_ls]
    data_ls = [line for line in data_ls if line != ""]
    return "\n".join(data_ls)


def parse_csv(data: str, delimeter: str = ",") -> list[list[str]]:
    """Parse CSV data."""

    data = remove_comments(data)
    data_ls = data.split("\n")
    data_ls_ls = [line.split(delimeter) for line in data_ls]
    data_ls_ls = remove_empty_items(data_ls_ls)
    data_ls_ls = [line for line in data_ls_ls if line != []]
    return data_ls_ls


def remove_empty_items(data: list[list[Any]]) -> list[list[Any]]:
    """Remove empty items from a list of lists."""

    data_ls: list[list[Any]] = []
    for line in data:
        line_ls: list[Any] = []
        for item in line:
            if item != "":
                line_ls.append(item)
        data_ls.append(line_ls)
    return data_ls



"""Handler for editting uncanny legends"""
from typing import Any

from . import event_stages
from ... import user_input_handler

def edit_uncanny(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editting uncanny legends"""
    stage_data = save_stats["uncanny"]
    lengths = stage_data["Lengths"]

    ids = []
    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter stage ids (e.g &1& = a new legend, &2& = here be dragons)(You can enter &all& to get all, a range e.g &1&-&49&, or ids separate by spaces e.g &5 4 7&):"
        ),
        lengths["total"],
    )
    save_stats["uncanny"] = event_stages.stage_handler(stage_data, ids, -1)

    return save_stats



"""Handler for basic, items that work in a common way"""
from typing import Any
from ... import helper, item, managed_item


def edit_cat_food(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing cat food"""

    cat_food = item.Item(
        name="Cat Food",
        value=save_stats["cat_food"]["Value"],
        max_value=45000,
        edit_name="value",
        bannable=item.BannableItem(
            is_bannable=True,
            has_workaround=False,
            managed_item_type=managed_item.ManagedItemType.CATFOOD,
        ),
    )
    cat_food.edit()
    save_stats["cat_food"]["Value"] = cat_food.value
    return save_stats


def edit_xp(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing xp"""

    experience = item.Item(
        name="XP",
        value=save_stats["xp"]["Value"],
        max_value=99999999,
        edit_name="value",
    )
    experience.edit()
    save_stats["xp"]["Value"] = experience.value
    return save_stats


def edit_normal_tickets(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing normal tickets"""

    normal_tickets = item.Item(
        name="Normal Tickets",
        value=save_stats["normal_tickets"]["Value"],
        max_value=2999,
        edit_name="value",
    )
    normal_tickets.edit()
    save_stats["normal_tickets"]["Value"] = normal_tickets.value
    return save_stats


def edit_rare_tickets(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing rare tickets"""

    rare_tickets = item.Item(
        name="Rare Tickets",
        value=save_stats["rare_tickets"]["Value"],
        max_value=299,
        edit_name="value",
        bannable=item.BannableItem(
            is_bannable=True,
            has_workaround=True,
            work_around_text='&Instead of editing rare tickets directly, use the "Normal Ticket Max Trade Progress" conversion feature instead! It is much more safe.',
            managed_item_type=managed_item.ManagedItemType.RARE_TICKET,
        ),
    )
    rare_tickets.edit()
    save_stats["rare_tickets"]["Value"] = rare_tickets.value
    return save_stats


def edit_platinum_tickets(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing platinum tickets"""

    platinum_tickets = item.Item(
        name="Platinum Tickets",
        value=save_stats["platinum_tickets"]["Value"],
        max_value=9,
        edit_name="value",
        bannable=item.BannableItem(
            is_bannable=True,
            has_workaround=True,
            work_around_text="&Instead of editing platinum tickets, edit platinum shards instead! They are much more safe. 10 platinum shards = 1 platinum ticket",
            managed_item_type=managed_item.ManagedItemType.PLATINUM_TICKET,
        ),
    )
    platinum_tickets.edit()
    save_stats["platinum_tickets"]["Value"] = platinum_tickets.value
    return save_stats


def edit_platinum_shards(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing platinum shards"""

    platinum_shards = item.Item(
        name="Platinum Shards",
        value=save_stats["platinum_shards"]["Value"],
        max_value=99,
        edit_name="value",
    )
    platinum_shards.edit()
    save_stats["platinum_shards"]["Value"] = platinum_shards.value
    return save_stats


def edit_np(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing np"""

    nyanko_points = item.Item(
        name="NP",
        value=save_stats["np"]["Value"],
        max_value=9999,
        edit_name="value",
    )
    nyanko_points.edit()
    save_stats["np"]["Value"] = nyanko_points.value
    return save_stats


def edit_leadership(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing leadership"""

    leadership = item.Item(
        name="Leadership",
        value=save_stats["leadership"]["Value"],
        max_value=9999,
        edit_name="value",
    )
    leadership.edit()
    save_stats["leadership"]["Value"] = leadership.value
    return save_stats


def edit_battle_items(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing battle items"""

    battle_items = item.create_item_group(
        names=[
            "Speed Up",
            "Treasure Radar",
            "Rich Cat",
            "Cat CPU",
            "Cat Jobs",
            "Sniper the Cat",
        ],
        values=save_stats["battle_items"],
        maxes=9999,
        edit_name="value",
        group_name="Battle Items",
    )
    battle_items.edit()
    save_stats["battle_items"] = battle_items.values

    return save_stats


def edit_catseyes(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing catseyes"""

    catseyes = item.create_item_group(
        names=[
            "Special",
            "Rare",
            "Super Rare",
            "Uber Super Rare",
            "Legend Rare",
        ],
        values=save_stats["catseyes"],
        maxes=9999,
        edit_name="value",
        group_name="Catseyes",
    )
    catseyes.edit()
    save_stats["catseyes"] = catseyes.values
    return save_stats


def edit_engineers(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing ototo engineers"""

    engineers = item.Item(
        name="Ototo Engineers",
        value=save_stats["engineers"]["Value"],
        max_value=5,
        edit_name="value",
    )
    engineers.edit()
    save_stats["engineers"]["Value"] = engineers.value
    return save_stats


def edit_base_materials(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing ototo base materials"""

    base_materials = item.create_item_group(
        names=[
            "Bricks",
            "Feathers",
            "Coal",
            "Sprockets",
            "Gold",
            "Meteorite",
            "Beast Bones",
            "Ammonite",
        ],
        values=save_stats["base_materials"],
        maxes=9999,
        edit_name="value",
        group_name="Base Materials",
    )
    base_materials.edit()
    save_stats["base_materials"] = base_materials.values
    return save_stats


def edit_catamins(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing catamins"""

    catamins = item.create_item_group(
        names=[
            "Catamin A",
            "Catamin B",
            "Catamin C",
        ],
        values=save_stats["catamins"],
        maxes=9999,
        edit_name="value",
        group_name="Catamins",
    )
    catamins.edit()
    save_stats["catamins"] = catamins.values
    return save_stats


def edit_inquiry_code(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing the inquiry code"""

    print(
        "WARNING: Editing your inquiry code should only be done if you know what you are doing! Because it will lead to an elsewhere error in-game if not done correctly!"
    )
    inquiry_code = item.Item(
        name="Inquiry Code",
        value=save_stats["inquiry_code"],
        max_value=None,
        edit_name="value",
    )
    inquiry_code.edit()
    save_stats["inquiry_code"] = inquiry_code.value
    return save_stats


def edit_rare_gacha_seed(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing the rare gacha seed"""

    rare_gacha_seed = item.Item(
        name="Rare Gacha Seed",
        value=save_stats["rare_gacha_seed"]["Value"],
        max_value=None,
        edit_name="value",
    )
    rare_gacha_seed.edit()
    save_stats["rare_gacha_seed"]["Value"] = rare_gacha_seed.value
    return save_stats


def edit_unlocked_slots(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing the amount of unlocked slots"""

    unlocked_slots = item.Item(
        name="Unlocked Slots",
        value=save_stats["unlocked_slots"]["Value"],
        max_value=len(save_stats["slots"]),
        edit_name="value",
    )
    unlocked_slots.edit()
    save_stats["unlocked_slots"]["Value"] = unlocked_slots.value
    return save_stats


def edit_token(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing the password-refresh-token"""

    print(
        "WARNING: Editing your token should only be done if you know what you are doing! Because it will lead to an elsewhere error in-game if not done correctly!"
    )
    token = item.Item(
        name="Token",
        value=save_stats["token"],
        max_value=None,
        edit_name="value",
    )
    token.edit()
    save_stats["token"] = token.value
    return save_stats


def edit_restart_pack(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for giving the restart pack"""

    save_stats["restart_pack"]["Value"] = 1
    print("Successfully gave the restart pack")
    return save_stats


def edit_challenge_battle(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing the score of the challenge battle"""

    challenge_battle = item.Item(
        name="Challenge Battle",
        value=save_stats["challenge"]["Score"]["Value"],
        max_value=None,
        edit_name="score",
    )
    challenge_battle.edit()
    save_stats["challenge"]["Score"]["Value"] = challenge_battle.value
    save_stats["challenge"]["Cleared"]["Value"] = 1
    return save_stats


def edit_legend_tickets(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing legend tickets"""

    legend_tickets = item.Item(
        name="Legend Tickets",
        value=save_stats["legend_tickets"]["Value"],
        max_value=4,
        edit_name="value",
        bannable=item.BannableItem(
            is_bannable=True,
            has_workaround=False,
            managed_item_type=managed_item.ManagedItemType.LEGEND_TICKET,
        ),
    )
    legend_tickets.edit()
    save_stats["legend_tickets"]["Value"] = legend_tickets.value
    return save_stats


def edit_dojo_score(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing the dojo score"""

    if not save_stats["dojo_data"]:
        helper.colored_text(
            "No catclaw dojo data found\nPlease enter the catclaw dojo menu and try again",
            base=helper.RED,
        )
        return save_stats

    dojo_score = item.Item(
        name="Dojo Score",
        value=save_stats["dojo_data"][0][0],
        max_value=None,
        edit_name="value",
    )
    dojo_score.edit()
    save_stats["dojo_data"][0][0] = dojo_score.value
    return save_stats



"""Handler for editing the aku realm"""
from typing import Any
from . import story_level_id_selector
from ... import helper, user_input_handler


def edit_aku(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Clear whole chapters"""

    options = ["Clear aku stages", "Wipe aku data"]
    option = user_input_handler.select_single(options, "select")
    if option == 2:
        return wipe_aku(save_stats)

    aku = save_stats["aku"]["Value"]

    ids = story_level_id_selector.select_levels(None, total=49)
    aku["clear_progress"][0][0] = max(ids)

    for level_id in ids:
        aku["clear_amount"][0][0][level_id] = 1
    save_stats["aku"]["Value"] = aku
    helper.colored_text("Successfully cleared aku stages")
    return save_stats


def wipe_aku(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Reset aku data"""

    aku = save_stats["aku"]["Value"]
    aku["clear_progress"][0][0] = 0
    aku["clear_amount"][0][0] = [0] * len(aku["clear_amount"][0][0])
    save_stats["aku"]["Value"] = aku
    helper.colored_text("Successfully wiped aku data")
    return save_stats



"""
Manager for config settings
"""
import os
from typing import Any

import yaml

from . import helper, user_input_handler


def get_config_value_category(category: str, key: str) -> Any:
    """
    Returns the value of the given key in the config file.
    """
    config = get_config_file()
    return config[category][key]


def get_config_file() -> dict[str, Any]:
    """
    Get the config file

    Returns:
        dict: Config file
    """
    config_file = os.path.join(get_app_data_folder(), "config.yaml")
    if not os.path.exists(config_file):
        create_config_file()
    with open(config_file, "r", encoding="utf-8") as file:
        config = yaml.safe_load(file)
    return config


def get_config_value(setting: str) -> Any:
    """
    Get a config setting

    Args:
        setting (str): Setting to get

    Returns:
        Any: Value of setting
    """
    config = get_config_file()
    return config[setting]


def set_config_setting_category(category: str, key: str, value: Any) -> None:
    """
    Set a config setting

    Args:
        setting (str): Setting to set
        value (Any): Value to set setting to
    """
    config = get_config_file()
    config[category][key] = value
    with open(
        os.path.join(get_app_data_folder(), "config.yaml"), "w", encoding="utf-8"
    ) as file:
        yaml.safe_dump(config, file)


def set_config_setting(setting: str, value: Any) -> None:
    """
    Set a config setting

    Args:
        setting (str): Setting to set
        value (Any): Value to set setting to
    """
    config = get_config_file()
    config[setting] = value
    with open(
        os.path.join(get_app_data_folder(), "config.yaml"), "w", encoding="utf-8"
    ) as file:
        yaml.safe_dump(config, file)


def create_config_file() -> None:
    """
    Create the config file if it doesn't exist
    """
    config_file = os.path.join(get_app_data_folder(), "config.yaml")
    file_data = "# Configuration file for BCSFE\n"
    file_data += "# This file is automatically created when the program is run for the first time\n"
    file_data += "# You can edit this file to change the settings\n"
    file_data += "# You can also edit the settings in the program\n#\n"
    file_data += "# The following settings are available:\n#\n"
    file_data += """DEFAULT_GAME_VERSION: "" # The default game version when downloading / pulling / loading save data. E.g en, jp, kr, tw. Empty means the game version is not specified and will be asked for when needed.
DEFAULT_SAVE_FILE_PATH: "" # The default file path for your save data when saving changes / downloading save data / pulling etc

EDITOR:
  DISABLE_MAXES: False # Allows you to edit the level / amount of items past the max amount.
  SHOW_BAN_WARNING: True # Show a warning when editing bannable items.
  SHOW_CATEGORIES: True # Show the categories in the feature list, instead of a long list.
  SHOW_FEATURE_SELECT_EXPLANATION: True # Show an explanation of how to select a feature.
  ONLY_GET_EN_DATA: False # Only get the en version of the game data even if the save is jp, use if you can't read japanese

START_UP:
  CHECK_FOR_UPDATES: True # Check for updates on startup
  UPDATE_TO_BETAS: False # Check for beta versions of the editor and update to them if found
  HIDE_START_TEXT: False # Hide the start up text - the credits, version info, other links, etc.
  DEFAULT_START_OPTION: -1 # The default save selection option when the editor starts, -1 gives you the option to do any option.
  CREATE_BACKUP: True # Create a backup of the save file when you start the editor.

SAVE_CHANGES:
  SAVE_CHANGES_ON_EDIT: False # Automatically save changes to your save data after using a feature.
  ALWAYS_EXPORT_JSON: False # Always export your save data as json when saving changes.

SERVER:
  UPLOAD_METADATA: True # Upload metadata (catfood, rare ticket, platinum ticket, and legend ticket changes) to the ponos servers when uploading and saving your save data - prevents bans.
  WIPE_TRACKED_ITEMS_ON_START: True # Wipe all tracked items stored when the editor starts up - if disabled, it allows you to upload metadata changes after you exit and re-enter the editor.
"""

    if not os.path.exists(config_file):
        helper.write_file_string(config_file, file_data)


def get_app_data_folder() -> str:
    """
    Get the path to the app data folder cross platform

    Returns:
        str: Path to app data folder
    """
    app_name = "BCSFE_Python"
    os_name = os.name
    if os_name == "nt":
        path = os.path.join(os.environ["APPDATA"], app_name)
    elif os_name == "mac":
        path = os.path.join(
            os.environ["HOME"], "Library", "Application Support", app_name
        )
    elif os_name == "posix":
        path = os.path.join(os.environ["HOME"], app_name)
    else:
        raise Exception(f"Unsupported platform {os_name}")
    helper.create_dirs(path)
    return path


def edit_default_gv(_: Any) -> None:
    """
    Edit the default game version
    """
    default_gv = get_config_value("DEFAULT_GAME_VERSION")
    default_gv = user_input_handler.colored_input(
        f"Enter the default game version. Current value: {default_gv} Leave blank to not specify a default game version:",
    )
    set_config_setting("DEFAULT_GAME_VERSION", default_gv)


def edit_default_save_file_path(_: Any) -> None:
    """
    Edit the default save file path
    """
    default_save_file_path = get_config_value("DEFAULT_SAVE_FILE_PATH")
    default_save_file_path = helper.select_dir(
        "Select the default save file path", default_save_file_path
    )
    set_config_setting(
        "DEFAULT_SAVE_FILE_PATH", os.path.join(default_save_file_path, "SAVE_DATA")
    )


def edit_editor_settings(_: Any) -> None:
    """
    Edit the editor settings
    """
    options = [
        "DISABLE_MAXES",
        "SHOW_BAN_WARNING",
        "SHOW_CATEGORIES",
        "SHOW_FEATURE_SELECT_EXPLANATION",
        "ONLY_GET_EN_DATA",
    ]
    option_values = [get_config_value_category("EDITOR", option) for option in options]
    ids = user_input_handler.select_not_inc(options, "select", option_values)
    for option_id in ids:
        option_name = options[option_id]
        enable = (
            user_input_handler.colored_input(
                f"Do you want to enable (&1&) or disable (&0&) {option_name}?:"
            )
            == "1"
        )
        set_config_setting_category("EDITOR", option_name, enable)


def edit_start_up_settings(_: Any) -> None:
    """
    Edit the start up settings

    Raises:
        Exception: If the option type is not int or bool
    """
    options = {
        "CHECK_FOR_UPDATES": bool,
        "UPDATE_TO_BETAS": bool,
        "HIDE_START_TEXT": bool,
        "DEFAULT_START_OPTION": int,
        "CREATE_BACKUP": bool,
    }
    option_values = [
        get_config_value_category("START_UP", option) for option in options
    ]
    ids = user_input_handler.select_not_inc(
        list(options.keys()), "select", option_values
    )
    for option_id in ids:
        option_name = list(options.keys())[option_id]
        option_type = options[option_name]
        if option_type == bool:
            enable = (
                user_input_handler.colored_input(
                    f"Do you want to enable (&1&) or disable (&0&) {option_name}?:"
                )
                == "1"
            )
            set_config_setting_category("START_UP", option_name, enable)
        elif option_type == int:
            option_value = user_input_handler.get_int(
                f"Enter the new value for {option_name}:"
            )
            set_config_setting_category("START_UP", option_name, option_value)
        else:
            raise Exception(f"Unsupported option type {option_type}")

def edit_save_changes_settings(_: Any) -> None:
    """
    Edit the save changes settings
    """
    options = [
        "SAVE_CHANGES_ON_EDIT",
        "ALWAYS_EXPORT_JSON",
    ]
    option_values = [get_config_value_category("SAVE_CHANGES", option) for option in options]
    ids = user_input_handler.select_not_inc(
        options, "select", option_values
    )
    for option_id in ids:
        option_name = options[option_id]
        enable = (
            user_input_handler.colored_input(
                f"Do you want to enable (&1&) or disable (&0&) {option_name}?:"
            )
            == "1"
        )
        set_config_setting_category("SAVE_CHANGES", option_name, enable)

def edit_server_settings(_: Any) -> None:
    """
    Edit the server settings
    """
    options = [
        "UPLOAD_METADATA",
        "WIPE_TRACKED_ITEMS_ON_START",
    ]
    option_values = [get_config_value_category("SERVER", option) for option in options]
    ids = user_input_handler.select_not_inc(
        options, "select", option_values
    )
    for option_id in ids:
        option_name = options[option_id]
        enable = (
            user_input_handler.colored_input(
                f"Do you want to enable (&1&) or disable (&0&) {option_name}?:"
            )
            == "1"
        )
        set_config_setting_category("SERVER", option_name, enable)


"""Handler for user input"""

from typing import Any, Tuple, Union

from . import helper


def handle_all_at_once(
    ids: list[int],
    all_at_once: bool,
    data: list[int],
    names: list[Any],
    item_name: str,
    group_name: str,
    explain_text: str = "",
) -> list[int]:
    """Handle all at once option"""

    first = True
    value = None
    for item_id in ids:
        if all_at_once and first:
            value = helper.check_int(
                colored_input(f"Enter {item_name} {explain_text}:")
            )
            first = False
        elif not all_at_once:
            value = helper.check_int(
                colored_input(
                    f"Enter {item_name} for {group_name} &{names[item_id]}& {explain_text}:"
                )
            )
        if value is None:
            continue
        data[item_id] = value
    return data


def create_all_list(
    ids: list[str],
    max_val: int,
) -> dict[str, Any]:
    """Creates a list with an all at once option"""

    all_at_once = False
    if f"{max_val}" in ids:
        ids_s = list(range(1, max_val))
        ids = [format(x, "02d") for x in ids_s]
        all_at_once = True
    return {"ids": ids, "at_once": all_at_once}


def create_all_list_inc(ids: list[str], max_val: int) -> dict[str, Any]:
    """Creates a list with an all at once option and include all"""

    return create_all_list(ids, max_val)


def create_all_list_not_inc(ids: list[str], max_val: int) -> list[str]:
    """Creates a list with an all at once option and don't include all"""

    return create_all_list(ids, max_val)["ids"]


def get_range(
    usr_input: str,
    length: Union[int, None] = None,
    min_val: int = 0,
    all_ids: Union[list[int], None] = None,
) -> list[int]:
    """Get a range of numbers from user input"""

    ids: list[int] = []
    for item in usr_input.split(" "):
        if item.lower() == "all":
            if length is None and all_ids is None:
                helper.colored_text("You can't use &all& here", helper.RED)
                return []
            if all_ids:
                return all_ids
            if length is not None:
                return list(range(min_val, length))
        if "-" in item:
            start_s, end_s = item.split("-")
            start = helper.check_int(start_s)
            end = helper.check_int(end_s)
            if start is None or end is None:
                helper.colored_text(
                    "Invalid input. Please enter a valid range of numbers separated by a dash.",
                    helper.RED,
                )
                return ids
            if start > end:
                start, end = end, start
            ids.extend(list(range(start, end + 1)))
        else:
            item_id = helper.check_int(item)
            if item_id is None:
                helper.colored_text(
                    "Invalid input. Please enter a valid integer.", helper.RED
                )
                return ids
            ids.append(item_id)
    return ids


def colored_input(dialog: str) -> str:
    """Format dialog as a colored string"""

    helper.colored_text(dialog, end="")
    return input()


def get_range_ids(group_name: str, length: int) -> list[int]:
    """Get a range of ids from user input"""

    ids = get_range(
        colored_input(
            f"Enter {group_name} ids(You can enter &all& to get all, a range e.g &1&-&50&, or ids separate by spaces e.g &5 4 7&):"
        ),
        length,
    )
    return ids


def select_options(
    options: list[str],
    mode: str = "edit",
    extra_data: Union[list[Any], None] = None,
    offset: int = 0,
) -> Tuple[list[int], bool]:
    """Select an option or multiple options from a list"""

    if len(options) == 1:
        return [0], True

    helper.colored_list(options, extra_data=extra_data, offset=offset)
    total = len(options)
    helper.colored_text(f"{total+1}. &Select all&")
    ids_s = colored_input(
        f"What do you want to {mode} (You can enter multiple values separated by spaces to {mode} multiple at once):"
    ).split(" ")
    individual = True
    if str(total + 1) in ids_s:
        ids = list(range(1, total + 1))
        individual = False
        ids_s = helper.int_to_str_ls(ids)

    ids = helper.parse_int_list(ids_s, -1)
    return ids, individual


def select_inc(
    options: list[str],
    mode: str = "edit",
    extra_data: Union[list[Any], None] = None,
    offset: int = 0,
) -> Tuple[list[int], bool]:
    """Select an option or multiple options from a list and include all"""

    return select_options(options, mode, extra_data, offset)


def select_not_inc(
    options: list[str],
    mode: str = "edit",
    extra_data: Union[list[Any], None] = None,
    offset: int = 0,
) -> list[int]:
    """Select an option or multiple options from a list and don't include all"""
    return select_options(options, mode, extra_data, offset)[0]


def select_single(
    options: list[str],
    mode: str = "edit",
) -> int:
    "Select a single option from a list"

    helper.colored_list(options)
    return get_int(f"What do you want to {mode}:")


def get_int(dialog: str) -> int:
    """Get user input as an integer and keep asking until a valid integer is entered"""

    helper.colored_text(dialog, end="")
    while True:
        try:
            val = input()
            val = val.strip(" ")
            return int(val)
        except ValueError:
            helper.colored_text(
                "Invalid input. Please enter a valid integer.", helper.RED
            )


def ask_if_individual(item_name: str) -> bool:
    """Ask if the user wants to edit an individual item"""

    is_individual = (
        colored_input(
            f"Do you want to edit {item_name} individually (&1&), or all at once (&2&):"
        )
        == "1"
    )
    return is_individual



"""Handler for getting the gold pass"""

import datetime
import time
from typing import Any
from ... import helper


def get_gold_pass_val(save_stats: dict[str, Any], total_days: int) -> dict[str, Any]:
    """Give the gold pass"""

    gold_pass = save_stats["gold_pass"]

    start_date = time.time()
    expiry_date = start_date + datetime.timedelta(days=total_days).total_seconds()
    gold_pass["officer_id"]["Value"] = 1
    gold_pass["flag"]["Value"] = 1
    gold_pass["start_date"] = start_date
    gold_pass["expiry_date"] = expiry_date
    gold_pass["start_date_2"] = start_date
    gold_pass["expiry_date_2"] = expiry_date
    gold_pass["unknown_3"] = start_date
    gold_pass["flag_2"]["Value"] = 1
    gold_pass["expiry_date_3"] = expiry_date

    save_stats["gold_pass"] = gold_pass

    return save_stats

def get_gold_pass(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Give the gold pass"""

    save_stats = get_gold_pass_val(save_stats, 30)

    helper.colored_text("Successfully gave the gold pass", helper.GREEN)

    return save_stats



"""Handler for evolving cats"""
from typing import Any

from ... import helper, csv_handler, game_data_getter
from . import cat_id_selector


def get_evolve(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for evolving cats"""

    cat_ids = cat_id_selector.select_cats(save_stats)
    return evolve_handler_ids(
        save_stats=save_stats,
        val=2,
        string="set",
        ids=cat_ids,
        forced=False,
    )


def get_evolve_forced(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for evolving cats without the form check"""

    cat_ids = cat_id_selector.select_cats(save_stats)
    return evolve_handler_ids(
        save_stats=save_stats,
        val=2,
        string="set",
        ids=cat_ids,
        forced=True,
    )


def remove_evolve(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for de-evolving cats"""

    cat_ids = cat_id_selector.select_cats(save_stats)
    return evolve_handler_ids(
        save_stats=save_stats,
        val=0,
        string="removed",
        ids=cat_ids,
        forced=True,
    )


def evolve_handler(
    save_stats: dict[str, Any], val: int, string: str, forced: bool
) -> dict[str, Any]:
    """Evolve specific cats"""

    ids = cat_id_selector.select_cats(save_stats)
    return evolve_handler_ids(save_stats, val, string, ids, forced)


def get_evolve_data(is_jp: bool) -> list[int]:
    """Get max form of cats"""

    data = helper.parse_int_list_list(
        csv_handler.parse_csv(
            game_data_getter.get_file_latest(
                "DataLocal", "nyankoPictureBookData.csv", is_jp
            ).decode("utf-8")
        )
    )
    forms = helper.copy_first_n(data, 2)
    forms = helper.offset_list(forms, -1)
    return forms


def get_total_forms(is_jp: bool, cat_id: int) -> int:
    """
    Get total forms of cat

    Args:
        is_jp (bool): Whether the game is JP or not
        cat_id (int): The cat id

    Returns:
        int: The total forms of the cat
    """
    data = get_evolve_data(is_jp)
    return data[cat_id]


def evolve_handler_ids(
    save_stats: dict[str, Any], val: int, string: str, ids: list[int], forced: bool
) -> dict[str, Any]:
    """Evolve specific cats by ids"""
    ids = helper.check_cat_ids(ids, save_stats)
    evolves = save_stats["unlocked_forms"]
    if not forced:
        form_data = get_evolve_data(helper.check_data_is_jp(save_stats))
        length = min([len(ids), len(form_data)])
        for i in range(length):
            evolves[ids[i]] = form_data[ids[i]]
    else:
        for cat_id in ids:
            evolves[cat_id] = val
    for cat_id, (unlocked_flag, current_flag) in enumerate(
        zip(evolves, save_stats["current_forms"])
    ):
        save_stats["current_forms"][cat_id] = max(unlocked_flag, current_flag)

    flags_evolved = [0 if form == 1 else form for form in evolves]
    save_stats["unlocked_forms"] = flags_evolved

    print(f"Successfully {string} true forms of cats")
    return save_stats



"""Handler for server requests"""

import datetime
import hashlib
import hmac
import json
from random import randint
import time
from typing import Any, Optional, Union

import requests

from . import helper, managed_item, serialise_save, patcher, config_manager, tracker


def get_current_time() -> int:
    """Get current time."""

    return int(time.time())


def random_hex_string(length: int) -> str:
    """Generate a random hex string."""

    return "".join([hex(randint(0, 15))[2:] for _ in range(length)])


def random_digit_string(length: int) -> str:
    """Generate a random digit string."""

    return "".join([str(randint(0, 9)) for _ in range(length)])


def get_nyanko_auth_url() -> str:
    """Get nyanko auth URL."""

    return "https://nyanko-auth.ponosgames.com"


def generate_nyanko_signature(inquiry_code: str, data: str) -> str:
    """Generate nyanko signature."""

    inquiry_code_bytes = inquiry_code.encode("utf-8")
    data_bytes = data.encode("utf-8")
    random_data = random_hex_string(64).encode("utf-8")
    input_rand = inquiry_code_bytes + random_data
    hmac_data = hmac.new(input_rand, data_bytes, digestmod=hashlib.sha256).hexdigest()

    final_signature = random_data.decode("utf-8") + hmac_data

    return final_signature


def generate_nyanko_signature_v1(inquiry_code: str, data: str) -> str:
    """Generate nyanko signature."""

    data += data

    inquiry_code_bytes = inquiry_code.encode("utf-8")
    data_bytes = data.encode("utf-8")
    random_data = random_hex_string(40).encode("utf-8")
    input_rand = inquiry_code_bytes + random_data
    hmac_data = hmac.new(input_rand, data_bytes, digestmod=hashlib.sha1).hexdigest()

    final_signature = random_data.decode("utf-8") + hmac_data

    return final_signature


def check_nyanko_signature(signature: str, data: str, inquiry_code: str) -> bool:
    """Check nyanko signature is correct"""

    curr_hmac_data = signature[64:]
    curr_random_data = signature[:64]
    curr_input_rand = inquiry_code.encode("utf-8") + curr_random_data.encode("utf-8")

    hmac_data = hmac.new(
        curr_input_rand, data.encode("utf-8"), digestmod=hashlib.sha256
    ).hexdigest()
    return hmac_data == curr_hmac_data


def create_backup_metadata(
    items: list[managed_item.ManagedItem],
    play_time: int,
    inquiry_code: str,
    user_rank: int,
) -> dict[str, Any]:
    """Create backup metadata."""
    managed_items: list[dict[str, Any]] = []
    for item in items:
        managed_items.append(item.to_dict())

    managed_items_s = json.dumps(managed_items)
    managed_items_s = managed_items_s.replace(" ", "")

    backup_metadata: dict[str, Any] = {
        "managedItemDetails": managed_items,
        "nonce": random_hex_string(32),
        "playTime": play_time,
        "rank": user_rank,
        "receiptLogIds": [],
        "signature_v1": generate_nyanko_signature_v1(inquiry_code, managed_items_s),
    }
    return backup_metadata


def check_nyanko_signature_v1(signature: str, data: str, inquiry_code: str) -> bool:
    """Check nyanko signature is correct"""

    data += data

    curr_hmac_data = signature[40:]
    curr_random_data = signature[:40]
    curr_input_rand = inquiry_code.encode("utf-8") + curr_random_data.encode("utf-8")

    hmac_data = hmac.new(
        curr_input_rand, data.encode("utf-8"), digestmod=hashlib.sha1
    ).hexdigest()
    return hmac_data == curr_hmac_data


def get_headers(inquiry_code: str, data: str) -> dict[str, Any]:
    """Get headers for a request."""

    return {
        "accept-encoding": "gzip",
        "connection": "keep-alive",
        "content-type": "application/json",
        "nyanko-signature": generate_nyanko_signature(inquiry_code, data),
        "nyanko-timestamp": str(get_current_time()),
        "nyanko-signature-version": "1",
        "nyanko-signature-algorithm": "HMACSHA256",
        "user-agent": "Dalvik/2.1.0 (Linux; U; Android 9; SM-G955F Build/N2G48B)",
    }


def handle_request(
    url: str, data: Union[str, bytes], headers: dict[str, Any]
) -> Union[dict[str, Any], None]:
    """Handle a request."""

    try:
        response = requests.post(url, data=data, headers=headers)
    except requests.exceptions.RequestException as err:
        raise Exception("Error getting password: " + str(err)) from err

    try:
        response_data = response.json()
    except json.decoder.JSONDecodeError as err:
        raise Exception("Error getting password: " + str(err)) from err

    if "statusCode" in response_data:
        if response_data["statusCode"] != 1:
            return None
        else:
            return response_data["payload"]
    else:
        return None


def get_password(inquiry_code: str) -> Union[dict[str, Any], None]:
    """Returns the account password for the given inquiry code."""

    url = get_nyanko_auth_url() + "/v1/users"
    data = {
        "accountCode": inquiry_code,
        "accountCreatedAt": str(
            get_current_time() - datetime.timedelta(days=1).seconds
        ),
        "nonce": random_hex_string(32),
    }

    data_s = json.dumps(data)
    data_s = data_s.replace(" ", "")
    headers = get_headers(inquiry_code, data_s)

    return handle_request(url, data_s, headers)


def get_password_refresh(
    inquiry_code: str, password_refresh_token: str
) -> Union[dict[str, Any], None]:
    """Returns the password for the given inquiry code"""

    url = get_nyanko_auth_url() + "/v1/user/password"
    data = {
        "accountCode": inquiry_code,
        "passwordRefreshToken": password_refresh_token,
        "nonce": random_hex_string(32),
    }
    data_s = json.dumps(data)
    data_s = data_s.replace(" ", "")
    headers = get_headers(inquiry_code, data_s)

    return handle_request(url, data_s, headers)


def get_client_info(country_code: str, game_version: str) -> dict[str, Any]:
    """Returns the client info for the given country_code and game_version"""

    country_code = country_code.replace("jp", "ja")

    data = {
        "clientInfo": {
            "client": {
                "countryCode": country_code,
                "version": game_version,
            },
            "device": {
                "model": "SM-G955F",
            },
            "os": {
                "type": "android",
                "version": "9",
            },
        },
        "nonce": random_hex_string(32),
    }
    return data


def get_token(
    inquiry_code: str, password: str, country_code: str, game_version: str
) -> Union[None, str]:
    """Returns the token for the given inquiry_code and password"""

    url = get_nyanko_auth_url() + "/v1/tokens"
    data = get_client_info(country_code, game_version)
    data["password"] = password
    data["accountCode"] = inquiry_code
    data_s = json.dumps(data)
    data_s = data_s.replace(" ", "")
    headers = get_headers(inquiry_code, data_s)

    payload = handle_request(url, data_s, headers)
    if payload is not None:
        return payload["token"]
    return None


def get_nyanko_save_url() -> str:
    """Get nyanko save URL."""

    return "https://nyanko-save.ponosgames.com"


def download_save(
    country_code: str,
    transfer_code: str,
    confirmation_code: str,
    game_version: str,
) -> bytes:
    """Downloads the save for the given country_code, transfer_code, confirmation_code
    and game_version"""

    country_code = country_code.replace("ja", "jp")
    url = get_nyanko_save_url() + "/v1/transfers/" + transfer_code + "/reception"
    data = get_client_info(country_code, game_version)
    data["pin"] = confirmation_code
    # data["isPasswordRefresh"] = True
    data_s = json.dumps(data)
    data_s = data_s.replace(" ", "")
    headers = {
        "content-type": "application/json",
        "accept-encoding": "gzip",
        "connection": "keep-alive",
        "user-agent": "Dalvik/2.1.0 (Linux; U; Android 9; SM-G955F Build/N2G48B)",
    }
    try:
        response = requests.post(url, data=data_s, headers=headers)
    except requests.exceptions.RequestException as err:
        raise Exception("Error getting save: " + str(err)) from err
    return response.content


def upload_save_data_body(
    managed_item_details: dict[str, Any],
    inquiry_code: str,
    token: str,
    save_data: bytes,
) -> tuple[bytes, dict[str, Any]]:
    """Gets the headers and body for uploading a save."""

    managed_item_details_str = json.dumps(managed_item_details)
    managed_item_details_str = managed_item_details_str.replace(" ", "")
    headers = get_headers(inquiry_code, managed_item_details_str)
    headers["authorization"] = "Bearer " + token

    boundary = f"__-----------------------{random_digit_string(9)}-2147483648".encode(
        "utf-8"
    )

    headers["content-type"] = "multipart/form-data; boundary=" + boundary.decode(
        "utf-8"
    )

    body = b"--" + boundary + b"\r\n"
    body += b'Content-Disposition: form-data; name="saveData"; filename="data.sav"\r\n'
    body += b"Content-Type: application/octet-stream\r\n\r\n"
    body += save_data + b"\r\n"
    body += b"--" + boundary + b"\r\n"
    body += b'Content-Disposition: form-data; name="metaData"\r\n\r\n'
    body += managed_item_details_str.encode("utf-8") + b"\r\n"
    body += b"--" + boundary + b"--\r\n"

    return body, headers


def upload_save_data(
    token: str,
    inquiry_code: str,
    save_data: bytes,
    play_time: int,
    items: list[managed_item.ManagedItem],
    user_rank: int,
):
    """Uploads the save data for the given token, inquiry_code and save_data"""

    if not config_manager.get_config_value_category("SERVER", "UPLOAD_METADATA"):
        items = []

    metadata = create_backup_metadata(items, play_time, inquiry_code, user_rank)
    body, headers = upload_save_data_body(metadata, inquiry_code, token, save_data)

    url = get_nyanko_save_url() + "/v1/transfers"
    return handle_request(url, body, headers)


def upload_metadata(
    token: str,
    inquiry_code: str,
    save_data: bytes,
    play_time: int,
    items: list[managed_item.ManagedItem],
    user_rank: int,
):
    """Uploads the metadata for the given token, inquiry_code and save_data"""

    metadata = create_backup_metadata(items, play_time, inquiry_code, user_rank)
    body, headers = upload_save_data_body(metadata, inquiry_code, token, save_data)

    url = get_nyanko_save_url() + "/v1/backups"
    return handle_request(url, body, headers)


def get_nyanko_backups_url() -> str:
    """Get nyanko backups URL."""

    return "https://nyanko-backups.ponosgames.com"


def get_nyanko_managed_items_url() -> str:
    """Get nyanko managed items URL."""

    return "https://nyanko-managed-item.ponosgames.com"


def get_inquiry_code() -> str:
    """Returns a new inquiry code"""

    url = get_nyanko_backups_url() + "/?action=createAccount&referenceId="
    response = requests.get(url)
    data = response.json()
    return data["accountId"]


def update_managed_items(
    inquiry_code: str, authorization: str, save_stats: dict[str, Any]
):
    """Updates the managed items"""
    # game should do this automatically, but eh why not just in case it doesn't

    data = {
        "catfoodAmount": save_stats["cat_food"]["Value"],
        "isPaid": False,
        "legendTicketAmount": save_stats["legend_tickets"]["Value"],
        "nonce": random_hex_string(32),
        "platinumTicketAmount": save_stats["platinum_tickets"]["Value"],
        "rareTicketAmount": save_stats["rare_tickets"]["Value"],
    }
    data_s = json.dumps(data)
    data_s = data_s.replace(" ", "")

    url = get_nyanko_managed_items_url() + "/v1/managed-items"
    headers = get_headers(inquiry_code, data_s)
    headers["authorization"] = "Bearer " + authorization
    return handle_request(url, data_s, headers)


def check_gen_token(
    save_stats: dict[str, Any], path: Optional[str] = None
) -> dict[str, Any]:
    """Gets the account auth token"""

    save_stats, password_refresh_data = check_gen_password(save_stats)
    inquiry_code = save_stats["inquiry_code"]

    save_data = None

    if path is not None:
        save_data = serialise_save.start_serialize(save_stats)
        save_data = patcher.patch_save_data(save_data, save_stats["version"])
        save_data = helper.write_file_bytes(path, save_data)

    token = get_token(
        inquiry_code,
        password_refresh_data["password"],
        save_stats["version"],
        save_stats["game_version"]["Value"],
    )

    return {
        "token": token,
        "save_data": save_data,
        "inquiry_code": inquiry_code,
        "password_refresh_data": password_refresh_data,
        "save_stats": save_stats,
    }


def check_gen_password(
    save_stats: dict[str, Any]
) -> tuple[dict[str, Any], dict[str, str]]:
    """Checks the password for the given save_stats and generates new if stuff is invalid"""

    inquiry_code = save_stats["inquiry_code"]
    password_refresh_token = save_stats["token"]

    password_refresh_data = get_password_refresh(inquiry_code, password_refresh_token)
    if password_refresh_data is not None:
        return save_stats, password_refresh_data

    password_refresh_data = get_password(inquiry_code)
    if password_refresh_data is None:
        inquiry_code = get_inquiry_code()
        save_stats["inquiry_code"] = inquiry_code

        return check_gen_password(save_stats)

    if "accountCode" in password_refresh_data:
        save_stats["inquiry_code"] = password_refresh_data["accountCode"]

    password_refresh_token = password_refresh_data["passwordRefreshToken"]
    save_stats["token"] = password_refresh_token
    return save_stats, password_refresh_data


def prepare_upload(
    save_stats: dict[str, Any], path: str, print_text: bool = True, managed_items: Optional[list[managed_item.ManagedItem]] = None
) -> Optional[tuple[str, Any, bytes, int, list[managed_item.ManagedItem]]]:
    """Handles the pre-upload of the save data"""

    original_iq = save_stats["inquiry_code"]
    if print_text:
        helper.colored_text("Getting account password...", helper.GREEN)
    data = check_gen_token(save_stats, path)
    token = data["token"]
    inquiry_code = data["inquiry_code"]
    save_data = data["save_data"]
    if token is None:
        if print_text:
            helper.colored_text(
                "Error getting account auth token",
                helper.RED,
            )
        return None
    item_tracker = tracker.Tracker()
    if original_iq != inquiry_code:
        item_tracker.reset_tracker()
        update_managed_items(save_stats["inquiry_code"], token, save_stats)

    if print_text:
        helper.colored_text("Adding meta data...", helper.GREEN)
    if managed_items is None:
        managed_items = item_tracker.parse_tracker_managed()
    playtime = helper.time_to_frames(save_stats["play_time"])
    tracker.Tracker().reset_tracker()

    return token, inquiry_code, save_data, playtime, managed_items


def upload_handler(
    save_stats: dict[str, Any], path: str
) -> Union[None, dict[str, Any]]:
    """Handles the upload of the save data"""

    data = prepare_upload(save_stats, path)
    if data is None:
        return None
    token, inquiry_code, save_data, playtime, managed_items = data

    helper.colored_text("Uploading save data...", helper.GREEN)
    upload_data = upload_save_data(
        token,
        inquiry_code,
        save_data,
        playtime,
        managed_items,
        helper.calculate_user_rank(save_stats),
    )

    helper.colored_text(
        "After entering these codes in game, you may get a ban message when pressing play. If you do, just press play again and it should go away.\nPress enter to get your codes...",
        helper.DARK_YELLOW,
    )
    input()
    return upload_data


def meta_data_upload_handler(
    save_stats: dict[str, Any], path: str
) -> tuple[Union[None, dict[str, Any]], dict[str, Any]]:
    """Handles the upload of the meta data"""

    data = prepare_upload(save_stats, path)
    if data is None:
        return None, save_stats
    token, inquiry_code, save_data, playtime, managed_items = data

    helper.colored_text("Uploading meta data...", helper.GREEN)
    upload_data = upload_metadata(
        token,
        inquiry_code,
        save_data,
        playtime,
        managed_items,
        helper.calculate_user_rank(save_stats),
    )

    return upload_data, save_stats

def test_is_save_data(save_data: bytes) -> bool:
    """Test if the save data is a valid save data"""

    try:
        save_data = json.loads(save_data)
    except (json.decoder.JSONDecodeError, UnicodeDecodeError):
        return True
    else:
        return False

def download_handler():
    """Handles the download of the save data"""

    country_code = helper.ask_cc()
    if country_code == "jp":
        country_code = "ja"
    transfer_code = helper.check_hex(
        input("Enter transfer code:").lower().replace("o", "0")
    )
    if transfer_code is None:
        helper.colored_text(
            "Transfer code must only be made up of numbers 0-9 and letters a-f",
            helper.RED,
        )
        return None
    confirmation_code = helper.check_dec(
        input("Enter confirmation code:").lower().replace("o", "0")
    )
    if confirmation_code is None:
        helper.colored_text(
            "Confirmation code must only be made up of numbers 0-9",
            helper.RED,
        )
        return None
    game_version = input("Enter current game version (e.g 11.3, 9.6, 10.4.0):")
    game_version = helper.str_to_gv(game_version)
    save_data = download_save(
        country_code, transfer_code, confirmation_code, game_version
    )
    if test_is_save_data(save_data):
        helper.colored_text("Successfully downloaded save data\n", base=helper.GREEN)
    else:
        helper.colored_text(
            "Incorrect transfer code / confirmation code",
            base=helper.RED,
        )
        return None
    path = helper.save_file(
        "Save save data",
        helper.get_save_file_filetype(),
        helper.get_default_save_name(),
    )
    helper.write_file_bytes(path, save_data)
    return path



"""Helper script for usefull functions"""

import json
from multiprocessing import Process
import os
import shutil
import sys
from tkinter import Tk, filedialog
from typing import Any, Generator, Union
import colored  # type: ignore

from . import (
    user_input_handler,
    server_handler,
    patcher,
    serialise_save,
    parse_save,
    tracker,
    config_manager,
)

GREEN = "#008000"
RED = "#FF0000"
DARK_YELLOW = "#D7C32A"
BLACK = "#000000"
WHITE = "#FFFFFF"
CYAN = "#00FFFF"


def print_line_seperator(base: str, char: str = "-", length: int = 80):
    """Print a line of a char"""
    width = shutil.get_terminal_size().columns
    if width < length:
        length = width
    colored_text(char * length, base)


def get_dirs(path: str) -> list[str]:
    """Get all directories in a path"""

    if not os.path.exists(path):
        return []

    return [dir for dir in os.listdir(path) if os.path.isdir(os.path.join(path, dir))]


def delete_dir(path: str) -> None:
    """Delete a directory and all of its contents"""

    for root, dirs, files in os.walk(path, topdown=False):
        for name in files:
            os.remove(os.path.join(root, name))
        for name in dirs:
            os.rmdir(os.path.join(root, name))
    os.rmdir(path)


def create_dirs(path: str) -> None:
    """Create directories if they don't exist"""

    if not os.path.exists(path):
        os.makedirs(path)


def offset_list(lst: list[int], offset: int) -> list[int]:
    """Offset each value in an list by a certain amount"""

    new_list: list[int] = []
    for item in lst:
        new_list.append(item + offset)
    return new_list


def copy_first_n(lst: list[Any], number: int) -> list[Any]:
    """Get the nth item in a list of lists"""

    new_list: list[Any] = []
    for item in lst:
        new_list.append(item[number])
    return new_list


def get_file(file_name: str) -> str:
    """Get file in files folder"""

    file_path = os.path.join(get_local_files_path(), file_name)
    return file_path


def find_files_in_dir(dir_path: str, file_name: str) -> list[str]:
    """Find all files in a directory with a certain name"""

    files: list[str] = []
    for file in os.listdir(dir_path):
        if file_name in file:
            files.append(file)
    return files


def get_local_files_path() -> str:
    """Get the local files path"""

    dir_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "files")
    return dir_path


def read_file_string(file_path: str) -> str:
    """Reads a file and returns its contents as a string"""

    try:
        with open(file_path, "r", encoding="utf-8") as file:
            return file.read()
    except FileNotFoundError as err:
        raise Exception("File not found: " + file_path) from err
    except UnicodeDecodeError as err:
        raise Exception("Error reading file: " + file_path + ": " + str(err)) from err


def chunks(lst: list[Any], chunk_len: int) -> Generator[Any, Any, Any]:
    """Split list into chunks of n"""

    for i in range(0, len(lst), chunk_len):
        yield lst[i : i + chunk_len]


def frames_to_time(frames: int) -> dict[str, Any]:
    """Turn frames into hours, minutes, seconds, frames"""

    seconds, frames = divmod(frames, 30)
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    return {"hh": hours, "mm": minutes, "ss": seconds, "frames": frames}


def num_to_bytes(num: int, length: int) -> bytes:
    """Turn number into little endian bytes"""

    return num.to_bytes(length, byteorder="little")


def seconds_to_time(seconds: int) -> dict[str, Any]:
    """Turn seconds into hours, minutes, seconds"""

    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    return {"hh": hours, "mm": minutes, "ss": seconds}


def time_to_seconds(time: dict[str, Any]) -> int:
    """Turn hours, minutes, seconds into seconds"""
    seconds = time["ss"]
    seconds += time["mm"] * 60
    seconds += time["hh"] * 60 * 60
    return seconds


def time_to_frames(time: dict[str, Any]) -> int:
    """Turn hours, minutes, seconds, frames into frames"""

    total_frames = time["frames"]
    total_frames += time["ss"] * 30
    total_frames += time["mm"] * 60 * 30
    total_frames += time["hh"] * 60 * 60 * 30
    return total_frames


def check_int(value: str) -> Union[int, None]:
    """Check if a string is an integer"""

    value = str(value).strip(" ")
    try:
        return int(value)
    except ValueError:
        return None


def int_to_str_ls(int_list: list[int]) -> list[str]:
    """Turn list of ints to list of strings"""

    str_list: list[str] = []
    for i in int_list:
        str_list.append(str(i))
    return str_list


def parse_int_list(lst: list[str], offset: int) -> list[int]:
    """Turn string list to int list"""

    new_list: list[int] = []
    for item in lst:
        try:
            new_list.append(int(item) + offset)
        except ValueError:
            pass
    return new_list


def clamp(value: int, min_value: int, max_value: int) -> int:
    """Clamp a value between two values"""

    return max(min(value, max_value), min_value)


def write_file_bytes(file_path: str, data: bytes) -> bytes:
    """Write file as bytes"""

    try:
        with open(file_path, "wb") as file:
            file.write(data)
    except PermissionError as err:
        raise Exception("Permission denied: " + file_path) from err
    return data


def get_save_path() -> str:
    """Get the save path from the env variable"""

    save_path = os.environ.get("BC_SAVE_PATH")
    if save_path is None:
        raise Exception("BC_SAVE_PATH not set")
    return save_path


def set_save_path(path: str) -> None:
    """Set the save path in the env variable"""

    os.environ["BC_SAVE_PATH"] = path


def get_text_splitter(isjp: bool):
    """Get the text splitter for the current save stats"""

    if isjp:
        return ","
    return "|"


def get_save_file_filetype() -> list[tuple[str, str]]:
    """Get the file types for the save file"""

    return [("Battle Cats Save Files", "*SAVE_DATA*"), ("All Files", "*.*")]


def read_file_bytes(file_path: str) -> bytes:
    """Read file as bytes"""

    with open(file_path, "rb") as file:
        return file.read()


def write_file_string(file_path: str, data: str):
    """Write file as string"""

    with open(file_path, "w", encoding="utf-8") as file:
        file.write(data)


def check_clamp(
    values: Any, max_value: int, min_value: int = 0, offset: int = -1
) -> list[int]:
    """turn a list of strings into a list of ints and clamp them between a min and max"""

    if isinstance(values, str):
        values = [values]
    int_values: list[int] = []
    for value in values:
        value = str(value).strip(" ")
        value = check_int(value)
        if value is None:
            continue
        value = clamp(value, min_value, max_value)
        value += offset
        int_values.append(value)
    return int_values


def encode_ls(lst: list[int]) -> dict[int, Any]:
    """Encode a list of integers into a dictionary"""
    return {i: lst[i] for i in range(len(lst))}


def parse_int_list_list(list_of_lists: list[list[str]]) -> list[list[Any]]:
    """Turn list of list of strings into list of list of ints"""
    new_lists: list[list[Any]] = []

    for lst in list_of_lists:
        new_list: list[Any] = []
        for item in lst:
            try:
                new_list.append(int(item))
            except ValueError:
                new_list.append(item)
        new_lists.append(new_list)
    new_lists = [line for line in new_lists if line != []]
    return new_lists


def check_hex(value: str) -> Union[str, None]:
    """Check if a string is a hex number"""

    value = str(value).strip(" ")
    try:
        int(value, 16)
        return value
    except ValueError:
        return None


def check_dec(value: str) -> Union[str, None]:
    """Check if a string is a decimal number"""

    value = str(value).strip(" ")
    try:
        int(value)
        return value
    except ValueError:
        return None


def str_to_gv(game_version: str) -> str:
    """Turn a game version with semantic versioning to integer representation"""

    split_gv = game_version.split(".")
    if len(split_gv) == 2:
        split_gv.append("0")
    final = ""
    for split in split_gv:
        final += split.zfill(2)

    return final.lstrip("0")


def gv_to_str(game_version: int) -> str:
    """Turn a game version with integer representation to semantic versioning"""

    split_gv = str(game_version).zfill(6)
    split_gv = [str(int(split_gv[i : i + 2])) for i in range(0, len(split_gv), 2)]
    return ".".join(split_gv)


def load_json(json_path: str) -> Any:
    """Load a json file"""

    return json.loads(read_file_string(json_path))


def is_jp(save_stats: dict[str, Any]) -> bool:
    """Check if the save is a Japanese save"""

    return save_stats["version"] == "jp"


def check_data_is_jp(save_stats: dict[str, Any]) -> bool:
    """Check if the save data is a Japanese save, checking the config file"""

    if config_manager.get_config_value_category("EDITOR", "ONLY_GET_EN_DATA"):
        return False
    return is_jp(save_stats)


def check_tracker(save_stats: dict[str, Any], path: str) -> None:
    """Check if the tracker is enabled"""
    item_tracker = tracker.Tracker()

    if item_tracker.has_data():
        upload = (
            user_input_handler.colored_input(
                "You have untracked bannable items that need to be uploaded. Do you want to upload them now? (&y&/&n&) (You will be unable to do so after exiting unless you have WIPE_TRACKED_ITEMS_ON_START set to False):"
            )
            == "y"
        )
        if upload:
            server_handler.meta_data_upload_handler(save_stats, path)
            colored_text("Uploaded meta data", new=GREEN)


def ask_exit_editor():
    """Exit the editor"""

    is_exit = (
        user_input_handler.colored_input("Do you want to exit the editor? (&y&/&n&):")
        == "y"
    )
    if is_exit:
        exit_editor()


def exit_editor():
    """Exit the editor"""

    sys.exit(0)


def check_cat_ids(cat_ids: list[int], save_stats: dict[str, Any]) -> list[int]:
    """Check if a list of cat ids is valid"""

    for cat_id in cat_ids:
        if cat_id > len(save_stats["cats"]):
            colored_text(f"Invalid cat id {cat_id}", base=RED)
            cat_ids.remove(cat_id)
    return cat_ids


def colored_text(
    text: str,
    base: str = WHITE,
    new: str = DARK_YELLOW,
    split_char: str = "&",
    end: str = "\n",
):
    """Print text with colors"""
    color_new = colored.fg(new)  # type: ignore
    color_base = colored.fg(base)  # type: ignore
    color_reset = colored.fg(WHITE)  # type: ignore

    text_split: list[str] = split_text(text, split_char)
    for i, text_section in enumerate(text_split):
        if i % 2:
            print(f"{color_new}{text_section}{color_base}", end="")
        else:
            print(f"{color_base}{text_section}{color_base}", end="")
    print(color_reset, end=end)


def split_text(text: str, split_char: str = "&") -> list[str]:
    """Split text on split_char, allowing for escaped split chars"""

    text_split: list[str] = []
    current_string = ""
    skip = 0
    for i, char in enumerate(text):
        if skip > 0:
            skip -= 1
            continue
        if char == "\\":
            if text[i + 1] == split_char:
                current_string += split_char
                skip = 1
                continue
        if char == split_char:
            text_split.append(current_string)
            current_string = ""
        else:
            current_string += char
    text_split.append(current_string)
    return text_split


def colored_list(
    items: list[str],
    extra_data: Any = None,
    index: bool = True,
    offset: Union[None, int] = None,
):
    """Print a list with colors and extra data if provided"""

    final = ""
    for i, item in enumerate(items):
        if index:
            final += f"{i+1}. "
        final += f"&{item}&"
        if extra_data:
            if extra_data[i] is not None:
                if isinstance(offset, int):
                    final += f" &:& {extra_data[i]+offset}"
                else:
                    final += f" &:& {extra_data[i]}"
        final += "\n"
    final = final.rstrip("\n")
    colored_text(final)


def calculate_user_rank(save_stats: dict[str, Any]):
    """Calculate the user rank"""

    user_rank = 10  # blue upgrade base level (1 each)
    user_rank += sum(save_stats["cats"])
    user_rank += sum(save_stats["cat_upgrades"]["Base"])
    user_rank += sum(save_stats["cat_upgrades"]["Plus"])
    user_rank += sum(save_stats["blue_upgrades"]["Base"])
    user_rank += sum(save_stats["blue_upgrades"]["Plus"])
    return user_rank


def write_save_data(save_data: bytes, country_code: str, path: str, prompt: bool):
    """Patch the save data and write it"""

    save_data = patcher.patch_save_data(save_data, country_code)
    if prompt:
        path = save_file("Save File", get_save_file_filetype(), path)
        if not path:
            colored_text("Save cancelled", new=RED)
            return
    write_file_bytes(path, save_data)
    colored_text(f"Saved to: &{os.path.abspath(path)}&", new=GREEN)
    return save_data


def select_dir(title: str, default_dir: str) -> str:
    """
    Select a directory from the user

    Args:
        title (str): Title of the dialog
        default_dir (str): Default directory to select

    Returns:
        str: Selected directory
    """
    root = setup_tk()
    dir_path = filedialog.askdirectory(title=title, initialdir=default_dir, parent=root)
    return dir_path


def setup_tk() -> Tk:
    """
    Setup the tkinter window

    Returns:
        Tk: Tkinter window
    """
    root = Tk()
    root.withdraw()
    root.wm_attributes("-topmost", 1)  # type: ignore
    return root


def run_in_parallel(fns: list[Process]) -> None:
    """
    Run a list of functions in parallel

    Args:
        fns (list[Process]): List of functions to run in parallel
    """
    proc: list[Process] = []
    for fn in fns:
        fn.start()
        proc.append(fn)
    for p in proc:
        p.join()


def get_cc(save_stats: dict[str, Any]) -> str:
    """Get the country code"""

    if is_jp(save_stats):
        return "jp"
    return "en"


def save_file(title: str, file_types: list[tuple[str, str]], path: str) -> str:
    """Save a file with tkinter"""
    setup_tk()

    try:
        path_d = filedialog.asksaveasfile(
            mode="w",
            confirmoverwrite=True,
            initialfile=os.path.basename(path),
            filetypes=file_types,
            title=title,
        )
        if not path_d:
            return ""
    except PermissionError:
        print(
            colored_text(
                "Permission denied. Make sure the file is not in use", base=RED
            )
        )
        exit_editor()
    if not path:
        return ""
    return path_d.name


def select_file(
    title: str,
    file_types: list[tuple[str, str]],
    default_dir: str = "",
    initial_file: str = "",
) -> str:
    """Select a file with tkinter"""
    setup_tk()
    file_path = filedialog.askopenfilename(
        initialdir=default_dir,
        title=title,
        filetypes=file_types,
        initialfile=initial_file,
    )
    return file_path


def get_default_save_name() -> str:
    """Get the default save name"""

    return config_manager.get_config_value("DEFAULT_SAVE_FILE_PATH")


def load_save_file(path: str) -> dict[str, Any]:
    """Load a save file, get the country code, create a backup and parse the save data"""

    save_data = read_file_bytes(path)
    country_code = get_country_code(save_data)
    colored_text(f"Game version: &{country_code}&")
    save_stats = parse_save.start_parse(save_data, country_code)
    if config_manager.get_config_value_category("START_UP", "CREATE_BACKUP"):
        write_file_bytes(path + "_backup", save_data)
        colored_text(
            f"Backup created at: &{os.path.abspath(path + '_backup')}&", new=GREEN
        )
    return {
        "save_data": save_data,
        "country_code": country_code,
        "save_stats": save_stats,
    }


def get_country_code(save_data: bytes) -> str:
    """Ask the user for their country code if it cannot be detected"""

    country_code = patcher.detect_game_version(save_data)
    if country_code is None:
        country_code = ask_cc()
    return country_code


def ask_cc():
    """Ask the user for their country code"""
    default_gv = config_manager.get_config_value("DEFAULT_GAME_VERSION")
    if default_gv:
        colored_text(f"Using default game version: &{default_gv}&")
        return default_gv

    country_code = user_input_handler.colored_input(
        "Enter your game version (&en&, &jp&, &kr&, &tw&):"
    )
    return country_code


def export_json(save_stats: dict[str, Any], path: str) -> None:
    """Export the save stats to a json file"""

    ordered_data = parse_save.re_order(save_stats)
    if os.path.isdir(path):
        path = os.path.join(path, f"{get_default_save_name()}.json")
    write_file_string(path, json.dumps(ordered_data, indent=4))
    colored_text(f"Successfully wrote json to &{os.path.abspath(path)}&")


def load_json_handler(json_path: str) -> Union[None, str]:
    """Load a save_data json file and serialise it"""

    save_stats = load_json(json_path)
    save_data = serialise_save.start_serialize(save_stats)
    save_data = patcher.patch_save_data(save_data, save_stats["version"])
    path = save_file(
        "Save file",
        get_save_file_filetype(),
        os.path.join(os.path.dirname(json_path), get_default_save_name()),
    )
    if not path:
        return None
    write_file_bytes(path, save_data)
    return path



"""Update the editor"""

import subprocess
from typing import Any

import requests

from . import config_manager, helper


def update(latest_version: str, command: str = "py") -> None:
    """Update pypi package testing for py and python"""

    helper.colored_text("Updating...", base=helper.GREEN)
    try:
        full_cmd = f"{command} -m pip install --upgrade battle-cats-save-editor=={latest_version}"
        subprocess.run(
            full_cmd,
            shell=True,
            capture_output=True,
            check=True,
        )
        helper.colored_text("Update successful", base=helper.GREEN)
    except subprocess.CalledProcessError as err:
        helper.colored_text("Update failed", base=helper.RED)
        if command == "py":
            helper.colored_text("Trying with python instead", base=helper.RED)
            update(latest_version, "python")
        else:
            helper.colored_text(
                f"Error: {err.stderr.decode('utf-8')}\nYou may need to manually update with py -m pip install -U battle-cats-save-editor",
                base=helper.RED,
            )


def get_local_version() -> str:
    """Returns the local version of the editor"""

    return helper.read_file_string(helper.get_file("version.txt"))


def get_version_info() -> tuple[str, str]:
    """Gets the latest version of the program"""

    package_name = "battle-cats-save-editor"
    try:
        response = requests.get(f"https://pypi.org/pypi/{package_name}/json")
        response.raise_for_status()
        data = response.json()
    except requests.exceptions.RequestException as err:
        raise Exception("Error getting pypi version") from err

    info = (
        get_pypi_version(data),
        get_latest_prerelease_version(data),
    )
    return info


def get_pypi_version(data: dict[str, Any]) -> str:
    """Get latest pypi version of the program"""
    return data["info"]["version"]


def get_latest_prerelease_version(data: dict[str, Any]) -> str:
    """Get latest prerelease version of the program"""
    releases = list(data["releases"])
    releases.reverse()
    for release in releases:
        if "b" in release:
            return release
    return ""


def pypi_is_newer(local_version: str, pypi_version: str, remove_b: bool = True) -> bool:
    """Checks if the local version is newer than the pypi version"""
    if remove_b:
        if "b" in pypi_version:
            pypi_version = pypi_version.split("b")[0]
        if "b" in local_version:
            local_version = local_version.split("b")[0]

    return pypi_version > local_version


def check_update(version_info: tuple[str, str]) -> tuple[bool, str]:
    """Checks if the editor is updated"""

    local_version = get_local_version()
    pypi_version, latest_prerelease_version = version_info

    check_pre = "b" in local_version or config_manager.get_config_value_category(
        "START_UP", "UPDATE_TO_BETAS"
    )
    if check_pre and pypi_is_newer(
        local_version, latest_prerelease_version, remove_b=False
    ):
        helper.colored_text("Prerelease update available\n", base=helper.GREEN)
        return True, latest_prerelease_version

    if pypi_is_newer(local_version, pypi_version):
        helper.colored_text("Stable update available\n", base=helper.GREEN)
        return True, pypi_version

    helper.colored_text("No update available\n", base=helper.GREEN)
    return False, local_version



"""Handler for creating a new account"""

from typing import Any

from . import fix_elsewhere
from ... import helper, server_handler


def create_new_account(save_stats: dict[str, Any]):
    """Create a new account"""

    helper.colored_text("Creating a new account...", helper.GREEN)

    save_stats["inquiry_code"] = server_handler.get_inquiry_code()
    save_stats["token"] = "0" * 40
    save_stats = fix_elsewhere.fix_elsewhere(save_stats, force_mi=True)

    return save_stats



"""Setup for installing the package."""

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("src/BCSFE_Python/files/version.txt", "r", encoding="utf-8") as fh:
    version = fh.read()

setuptools.setup(
    name="battle-cats-save-editor",
    version=version,
    author="fieryhenry",
    description="A battle cats save file editor",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fieryhenry/BCSFE-Python",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.9",
    install_requires=["colored", "tk", "python-dateutil", "requests", "pyyaml"],
    include_package_data=True,
    extras_require={
        "testing": [
            "pytest",
            "pytest-cov",
        ],
    },
    package_data={"BCSFE_Python": ["py.typed"]},
    flake8 = {"max-line-length": 160},
)



"""Handler for selecting cat ids"""

from multiprocessing import Process
import os
from typing import Any, Callable, Optional

from ... import user_input_handler, helper, game_data_getter, csv_handler
from . import upgrade_cats
from ..levels import treasures


def select_cats(save_stats: dict[str, Any], current: bool = True) -> list[int]:
    """Select cats"""

    options: dict[str, Callable[[dict[str, Any]], list[int]]] = {
        "Select currently unlocked cats" : select_current_cats,
        "Select cats of a certain rarity" : select_cats_rarity,
        "Select specific cat ids" : select_cats_range,
        "Select cats of a specific gacha banner": select_cats_gatya_banner,
        "Select all cats": get_all_cats,
        "Search by cat name": select_cat_names,
        "Select all obtainable cats": select_cats_obtainable,
    }
    if not current:
        del options["Select currently unlocked cats"]
    choice = user_input_handler.select_single(list(options), "select") - 1
    cat_ids = options[list(options)[choice]](save_stats)
    return cat_ids

def select_cats_obtainable(save_stats: dict[str, Any]) -> list[int]:
    """
    Select cats that can be obtained

    Args:
        save_stats (dict[str, Any]): Save stats

    Returns:
        list[int]: Cat ids
    """
    return filter_obtainable_cats(save_stats, get_all_cats(save_stats))

def select_current_cats(save_stats: dict[str, Any]) -> list[int]:
    """Select current cats"""

    cats = save_stats["cats"]
    cat_ids: list[int] = []
    for i, cat_val in enumerate(cats):
        if cat_val == 1:
            cat_ids.append(i)
    return cat_ids


def select_cats_rarity(save_stats: dict[str, Any]) -> list[int]:
    """Select cats of a certain rarity"""

    ids = user_input_handler.select_not_inc(
        options=upgrade_cats.TYPES,
        mode="upgrade",
    )
    is_jp = helper.is_jp(save_stats)

    cat_ids = upgrade_cats.get_rarity(ids, is_jp)
    return cat_ids


def select_cats_range(save_stats: dict[str, Any]) -> list[int]:
    """Select cats in a range"""

    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter cat ids (Look up cro battle cats to find ids)(You can enter &all& to get all, a range e.g &1&-&50&, or ids separate by spaces e.g &5 4 7&):"
        ),
        length=len(save_stats["cats"]),
    )
    return ids


def select_cats_gatya_banner(save_stats: dict[str, Any]) -> list[int]:
    """Select cats for a specific gacha banner"""
    is_jp = helper.is_jp(save_stats)
    data = helper.parse_int_list_list(
        csv_handler.parse_csv(
            game_data_getter.get_file_latest(
                "DataLocal", "GatyaDataSetR1.csv", is_jp
            ).decode("utf-8")
        )
    )
    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter gacha banner id (Look up the gacha banners you want, then click on the image at the top, and look for the last digits of the file name (e.g royal fest = 602)):"
        ),
        length=len(data),
    )
    data = treasures.remove_negative_1(data)
    cat_ids: list[int] = []
    for c_id in ids:
        cat_ids.extend(data[c_id])
    return cat_ids


def get_all_cats(save_stats: dict[str, Any]) -> list[int]:
    """Get all cats"""

    return list(range(len(save_stats["cats"])))


def select_cat_names(save_stats: dict[str, Any]) -> list[int]:
    """
    select_cat_names

    Args:
        save_stats (dict[str, Any]): save stats

    Returns:
        list[int]: cat ids
    """
    all_names = get_cat_names(save_stats)
    name = user_input_handler.colored_input("Enter cat name:")
    found_names = search_cat_names(name, all_names)
    found_names = filter_cat_names(found_names)
    if not found_names:
        print("No cats with that name found")
        return []

    cat_ids: list[int] = []
    cat_names: list[str] = []
    for cat_name, cat_id, _ in found_names:
        cat_ids.append(cat_id)
        cat_name = cat_name.replace("&", "\\&")
        cat_names.append(cat_name)

    indexes = user_input_handler.select_not_inc(
        cat_names, mode="select", extra_data=cat_ids
    )
    cat_ids = [cat_ids[i] for i in indexes]
    return cat_ids


def get_cat_by_form_and_id(
    all_names: list[tuple[str, int, int]], cat_id: int, form_id: int
) -> Optional[tuple[str, int, int]]:
    """
    Get cat by form and id

    Args:
        all_names (list[tuple[str, int, int]]): all names
        cat_id (int): cat id
        form_id (int): form id

    Returns:
        Optional[tuple[str, int, int]]: cat data
    """
    for cat in all_names:
        if cat[1] == cat_id and cat[2] == form_id:
            return cat
    return None


def get_cat_by_id(
    cat_names: list[tuple[str, int, int]], cat_id_to_search: int
) -> list[tuple[str, int, int]]:
    """
    Get cat by id

    Args:
        cat_names (list[tuple[str, int, int]]): list of cat names
        cat_id_to_search (int): cat id to search for

    Returns:
        Optional[tuple[str, int, int]]: cat name, cat id, cat form
    """
    cats: list[tuple[str, int, int]] = []
    for cat_name, cat_id, cat_form in cat_names:
        if cat_id == cat_id_to_search:
            cats.append((cat_name, cat_id, cat_form))
    return cats


def filter_cat_names(
    cat_names: list[tuple[str, int, int]]
) -> list[tuple[str, int, int]]:
    """
    Filter cat names by only selecting one of the forms

    Args:
        cat_names (list[tuple[str, int, int]]): list of cat names

    Returns:
        list[tuple[str, int, int]]: filtered cat names
    """
    filtered_cat_ids: list[int] = []
    cat_data: list[tuple[str, int, int]] = []
    for cat_name, cat_id, cat_form in cat_names:
        if cat_id not in filtered_cat_ids:
            filtered_cat_ids.append(cat_id)
            cat_data.append((cat_name, cat_id, cat_form))

    return cat_data


def search_cat_names(
    name: str, cat_names: list[tuple[str, int, int]]
) -> list[tuple[str, int, int]]:
    """
    Search cat names

    Args:
        name (str): name to search for
        cat_names (list[tuple[str, int, int]]): list of cat names

    Returns:
        list[tuple[str, int, int]]: list of cat names that match the search
    """

    found_names: list[tuple[str, int, int]] = []

    for cat_name, cat_id, form_id in cat_names:
        if name.lower().replace(" ", "") in cat_name.lower().replace(" ", ""):
            found_names.append((cat_name, cat_id, form_id))
    return found_names


def download_10_files(game_version: str, file_names: list[str]) -> None:
    """
    Download 10 files

    Args:
        game_version (str): game version
        file_names (list[str]): file names
    """
    for file_name in file_names:
        game_data_getter.download_file(game_version, "resLocal", file_name, False)


def get_cat_names(save_stats: dict[str, Any]) -> list[tuple[str, int, int]]:
    """
    Get cat names and ids

    Args:
        save_stats (dict[str, Any]): save stats

    Returns:
        list[tuple[str, int, int]]: cat names and ids
    """

    is_jp = helper.is_jp(save_stats)

    file_path_dir = os.path.dirname(
        helper.get_file(game_data_getter.get_path("resLocal", "", is_jp))
    )
    helper.create_dirs(file_path_dir)
    if len(helper.find_files_in_dir(file_path_dir, "Unit_Explanation")) < len(
        save_stats["cats"]
    ):
        helper.colored_text(
            "Downloading cat names for the first time... (This may take some time, but next time it will be much faster)",
            helper.GREEN,
        )
        funcs: list[Process] = []
        version = game_data_getter.get_latest_version(is_jp)
        all_file_names: list[str] = []
        for cat_id, _ in enumerate(save_stats["cats"]):
            file_name = f"Unit_Explanation{cat_id+1}_{helper.get_cc(save_stats)}.csv"
            all_file_names.append(file_name)
        file_names_split = helper.chunks(all_file_names, 10)
        for file_names in file_names_split:
            funcs.append(
                Process(
                    target=download_10_files,
                    args=(version, file_names),
                )
            )
        helper.run_in_parallel(funcs)

    names: list[tuple[str, int, int]] = []
    for cat_id, _ in enumerate(save_stats["cats"]):
        file_path = os.path.join(
            file_path_dir, f"Unit_Explanation{cat_id+1}_{helper.get_cc(save_stats)}.csv"
        )
        data = csv_handler.parse_csv(
            helper.read_file_string(file_path),
            delimeter=helper.get_text_splitter(is_jp),
        )
        for form_id, form in enumerate(data):
            name = form[0]
            names.append((name, cat_id, form_id))
    return names


def get_obtainability(save_stats: dict[str, Any]) -> list[int]:
    """
    Get obtainability of cats

    Args:
        save_stats (dict[str, Any]): save stats

    Returns:
        list[int]: obtainability of cats (0 = not obtainable, 1 = obtainable)
    """
    data = helper.parse_int_list_list(
        csv_handler.parse_csv(
            game_data_getter.get_file_latest(
                "DataLocal", "nyankoPictureBookData.csv", helper.is_jp(save_stats)
            ).decode("utf-8")
        )
    )
    is_obtainable = helper.copy_first_n(data, 0)
    return is_obtainable


def get_obtainable_cats(save_stats: dict[str, Any]) -> list[int]:
    """
    Get obtainable cats

    Args:
        save_stats (dict[str, Any]): save stats

    Returns:
        list[int]: obtainable cats
    """
    obtainability = get_obtainability(save_stats)
    return [i for i, x in enumerate(obtainability) if x == 1]


def filter_obtainable_cats(save_stats: dict[str, Any], cat_ids: list[int]) -> list[int]:
    """
    Filter obtainable cats in a list of cat ids

    Args:
        save_stats (dict[str, Any]): save stats
        cat_ids (list[int]): cat ids

    Returns:
        list[int]: obtainable cats
    """
    obtainable_cats = get_obtainable_cats(save_stats)
    return [i for i in cat_ids if i in obtainable_cats]



"""Handle adb commands"""

import enum
import os
import subprocess
from typing import Union
from . import helper, user_input_handler


class ADBExceptionTypes(enum.Enum):
    """ADB exception types"""

    NO_DEVICE = enum.auto()
    DEVICE_OFFLINE = enum.auto()
    PATH_NOT_FOUND = enum.auto()
    ADB_NOT_INSTALLED = enum.auto()
    MORE_THAN_ONE_DEVICE = enum.auto()
    UNKNOWN = enum.auto()


class ADBException(Exception):
    """ADB exception"""

    def __init__(self, exception_type: ADBExceptionTypes, message: str = ""):
        """Initialize exception."""

        super().__init__(message)
        self.message = message
        self.exception_type = exception_type


def adb_pull(package_name: str, device_file_path: str, local_file_path: str):
    """Pull a file from a device"""
    if local_file_path:
        local_file_path = f'"{local_file_path}"'
    path = f"/data/data/{package_name}/{device_file_path}"
    run_adb_command(f'pull "{path}" {local_file_path}')


def adb_push(package_name: str, local_file_path: str, device_file_path: str):
    """Push a file to a device"""

    path = f"/data/data/{package_name}/{device_file_path}"
    run_adb_command(f'push "{local_file_path}" "{path}"')


def adb_delete_file(package_name: str, device_file_path: str, options: str = ""):
    """Delete a file on a device"""

    path = f"/data/data/{package_name}/{device_file_path}"
    run_adb_command(f'shell "su 0 rm "{path}" {options}"')


def adb_close_process(package_name: str):
    """Close a process"""

    run_adb_command(f"shell am force-stop {package_name}")


def adb_run_process(package_name: str):
    """Run a process"""

    run_adb_command(f"shell monkey -p {package_name} -v 1")

def adb_reboot():
    """Reboot adb server"""

    adb_kill_server()
    is_adb_installed()

def adb_root():
    """Start adb server as root"""

    subprocess.run("adb root", shell=True, check=True, text=True, capture_output=True)


def is_adb_installed():
    """Test if adb is installed"""

    try:
        subprocess.run(
            "adb start-server",
            shell=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=True,
        )
    except subprocess.CalledProcessError:
        return False
    return True


def run_adb_command(command: str) -> bool:
    """Run an ADB command"""

    command = f"adb {command}"
    if not is_adb_installed():
        raise ADBException(ADBExceptionTypes.ADB_NOT_INSTALLED)
    try:
        adb_root()
        subprocess.run(command, shell=True, check=True, text=True, capture_output=True)
    except subprocess.CalledProcessError as err:
        adb_error_handler(err)
    return True


def adb_kill_server():
    """Kill ADB server"""

    try:
        subprocess.run("adb kill-server", shell=True, check=True, text=True, capture_output=True)
    except subprocess.CalledProcessError as err:
        adb_error_handler(err)


def adb_error_handler(err: subprocess.CalledProcessError):
    """Handle ADB errors"""

    if "not found" in err.stderr:
        raise ADBException(ADBExceptionTypes.NO_DEVICE)
    if "offline" in err.stderr:
        print(err.stderr)
        raise ADBException(ADBExceptionTypes.DEVICE_OFFLINE)
    if "does not exist" in err.stderr:
        raise ADBException(ADBExceptionTypes.PATH_NOT_FOUND)
    if "'adb' is not recognized" in err.stderr:
        raise ADBException(ADBExceptionTypes.ADB_NOT_INSTALLED)
    if "more than one device" in err.stderr:
        raise ADBException(ADBExceptionTypes.MORE_THAN_ONE_DEVICE)
    raise ADBException(ADBExceptionTypes.UNKNOWN, err.stderr)


def find_adb_path() -> Union[str, None]:
    """Find adb path automatically in common locations"""

    drive_letters = ["C", "D", "E"]
    paths = [
        "LDPlayer\\LDPlayer4.0",
        "LDPlayer\\LDPlayer9",
        "Program Files (x86)\\Nox\\bin",
        "adb",
    ]
    for drive_letter in drive_letters:
        for path in paths:
            path = f"{drive_letter}:\\{path}"
            if os.path.exists(path):
                return path

    return None

def if_windows() -> bool:
    """Check if windows"""

    return os.name == "nt"

def add_to_path() -> None:
    """Try to add adb to path environment variable automatically"""

    if not if_windows():
        helper.colored_text(
            "ADB path not added to PATH environment variable.\n"
            "Please add it manually to your system PATH variable.",
            helper.RED,
        )
        return

    adb_path = find_adb_path()
    if not adb_path:
        adb_path = input(
            "Please enter the path to the folder than contains adb: download link here: https://dl.google.com/android/repository/platform-tools-latest-windows.zip:"
        )
        if os.path.isfile(adb_path):
            adb_path = os.path.dirname(adb_path)

    print(f"Adding {adb_path} to your path environment variable")
    subprocess.run(f"set PATH={adb_path};%PATH%", shell=True, check=True)
    subprocess.run(f"setx PATH {adb_path};%PATH%", shell=True, check=True)
    print("Successfully added adb to path")


def adb_err_handler(err: ADBException):
    """Handle ADB errors"""
    if err.exception_type in (
        ADBExceptionTypes.NO_DEVICE,
        ADBExceptionTypes.DEVICE_OFFLINE,
    ):
        helper.colored_text(
            "Error: No device with an adb connection can be found, please connect one and try again. (You may have to wait aprox 1min for the device to be detected)",
            base=helper.RED,
        )
        adb_reboot()
    elif err.exception_type == ADBExceptionTypes.PATH_NOT_FOUND:
        helper.colored_text(
            "Error: You don't seem to have this game version installed on this device / no root access / SAVE_DATA couldn't be located, please make sure you have loaded into the game and clicked \"START\" and try again.",
            base=helper.RED,
        )
    elif err.exception_type == ADBExceptionTypes.ADB_NOT_INSTALLED:
        add_adb = (
            user_input_handler.colored_input(
                "Error, adb is not in your Path environment variable. There is a tutorial in the github's readme. Would you like to try to add adb to your path now?(&y&/&n&):"
            )
            == "y"
        )
        if add_adb:
            add_to_path()
            print("Please re-run the editor to try again")
    elif err.exception_type == ADBExceptionTypes.MORE_THAN_ONE_DEVICE:
        helper.colored_text(
            "Error: More than one device with an adb connection can be found, please make sure that only 1 device is connected. (You may have to wait aprox 1min for the device to be detected)",
            base=helper.RED,
        )
        adb_reboot()
    else:
        helper.colored_text(
            "Error: " + str(err.message),
            base=helper.RED,
        )
    helper.exit_editor()


def adb_pull_save_data(game_version: str) -> str:
    """Pull save data from device"""
    helper.colored_text(
        "Pulling save data from device...",
        base=helper.DARK_YELLOW,
    )
    try:
        adb_pull(
            get_package_name(game_version),
            "files/SAVE_DATA",
            helper.get_default_save_name(),
        )
    except ADBException as err:
        adb_err_handler(err)
    return helper.get_default_save_name()


def adb_push_save_data(game_version: str, path: str) -> None:
    """Push save data to device"""
    helper.colored_text(
        "Pushing save data to device...",
        base=helper.DARK_YELLOW,
    )
    try:
        adb_push(get_package_name(game_version), path, "files/SAVE_DATA")
    except ADBException as err:
        adb_err_handler(err)


def rerun_game(game_version: str) -> None:
    """Rerun game"""

    helper.colored_text(
        "Rerunning game...",
        base=helper.DARK_YELLOW,
    )
    try:
        adb_close_process(get_package_name(game_version))
        adb_run_process(get_package_name(game_version))
    except ADBException as err:
        adb_err_handler(err)


def adb_clear_save_data(game_version: str) -> None:
    """Clear save data"""

    try:
        adb_delete_file(get_package_name(game_version), "/files/*SAVE_DATA*")
        adb_delete_file(get_package_name(game_version), "/shared_prefs", "-r -f")
    except ADBException as err:
        adb_err_handler(err)


def get_package_name(game_version: str) -> str:
    """Get package name"""

    if game_version == "jp":
        game_version = ""
    return f"jp.co.ponos.battlecats{game_version}"



"""Handler for editing the ototo cat cannon"""
from typing import Any

from ... import user_input_handler, item, game_data_getter, csv_handler, helper


def get_canon_types(is_jp: bool):
    """Get the cannon types"""

    data = csv_handler.parse_csv(
        game_data_getter.get_file_latest(
            "resLocal", "CastleRecipeDescriptions.csv", is_jp
        ).decode("utf-8"),
        delimeter=helper.get_text_splitter(is_jp),
    )
    types: list[str] = []
    for cannon in data:
        types.append(cannon[1])
    return types


def get_cannon_maxes(is_jp: bool) -> list[int]:
    """Get the cannon maxes"""

    data = helper.parse_int_list_list(
        csv_handler.parse_csv(
            game_data_getter.get_file_latest(
                "DataLocal", "CastleRecipeUnlock.csv", is_jp
            ).decode("utf-8")
        )
    )
    maxes: list[int] = []
    for cannon in data:
        cannon_id = cannon[0]
        max_val = cannon[3] - 1
        if cannon_id == len(maxes):
            maxes.append(max_val)
        else:
            if maxes[cannon_id] < max_val:
                maxes[cannon_id] = max_val
    return maxes


def set_level_val(
    cannons: list[dict[str, Any]], levels: list[int]
) -> list[dict[str, Any]]:
    """Set the upgrade level of the cannon"""

    for i, level in enumerate(levels):
        if level > 0:
            cannons[i]["unlock_flag"] = 3
        cannons[i]["level"] = level
    return cannons


def set_stage_val(
    cannons: list[dict[str, Any]], stages: list[int]
) -> list[dict[str, Any]]:
    """Set the stage of the cannon development"""

    for i, stage in enumerate(stages):
        cannons[i + 1]["unlock_flag"] = stage
    return cannons


def get_data(cannons: dict[int, dict[str, int]]) -> tuple[list[int], list[int]]:
    """Get the data of the cannon"""

    levels: list[int] = []
    stages: list[int] = []
    for cannon in cannons.values():
        level = cannon["level"]
        stage = cannon["unlock_flag"]
        levels.append(level)
        stages.append(stage)
    return levels, stages


def set_level(save_stats: dict[str, Any], levels: list[int]) -> dict[str, Any]:
    """Set the upgrade level of the cannon"""

    cannons = save_stats["ototo_cannon"]

    ot_levels = item.create_item_group(
        names=get_canon_types(helper.check_data_is_jp(save_stats)),
        values=levels,
        maxes=get_cannon_maxes(helper.check_data_is_jp(save_stats)),
        edit_name="level",
        group_name="Cannon Level",
        offset=1,
    )
    ot_levels.edit()

    save_stats["ototo_cannon"] = set_level_val(cannons, ot_levels.values)
    return save_stats


def set_stage(save_stats: dict[str, Any], stages: list[int]) -> dict[str, Any]:
    """Set the stage of the cannon development"""

    cannons = save_stats["ototo_cannon"]

    ot_stages = item.create_item_group(
        names=get_canon_types(helper.check_data_is_jp(save_stats))[1:],
        values=stages[1:],
        maxes=3,
        edit_name="stage",
        group_name="Ototo Cat Cannon Stage",
    )
    ot_stages.edit()
    save_stats["ototo_cannon"] = set_stage_val(cannons, ot_stages.values)
    return save_stats


def edit_cat_cannon(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for ototo cat cannon upgrades"""

    cannons = save_stats["ototo_cannon"]
    levels, stages = get_data(cannons)
    stage = (
        user_input_handler.colored_input(
            "Do you want to set the level of the cannon &(1)& or the level of construction &(2)& (e.g foundation, style, cannon):"
        )
        == "2"
    )
    if stage:
        save_stats = set_stage(save_stats, stages)
    else:
        save_stats = set_level(save_stats, levels)
    return save_stats



"""Handler for editing gamatoto xp"""
from typing import Any

from ... import helper, user_input_handler, item, game_data_getter

def get_boundaries(is_jp: bool) -> list[int]:
    """Get the xp requirements for each level"""

    boundaries = game_data_getter.get_file_latest(
        "DataLocal", "GamatotoExpedition.csv", is_jp
    ).decode("utf-8").splitlines()
    previous = 0
    xp_requirements: list[int] = []
    previous = 0
    for line in boundaries:
        requirement = int(line.split(",")[0])
        if previous >= requirement:
            break
        xp_requirements.append(requirement)
        previous = requirement
    return xp_requirements


def get_level_from_xp(gamatoto_xp: int, is_jp: bool) -> dict[str, Any]:
    """Get the level from the xp amount"""

    xp_requirements = get_boundaries(is_jp)
    level = 1
    for requirement in xp_requirements:
        if gamatoto_xp >= requirement:
            level += 1
    return {
        "level": level,
        "max_level": len(xp_requirements),
        "max_xp": xp_requirements[-2],
    }


def get_xp_from_level(level: int, is_jp: bool) -> int:
    """Get the xp amount from the level"""

    xp_requirements = get_boundaries(is_jp)
    if level <= 1:
        gamatoto_xp = 0
    else:
        gamatoto_xp = xp_requirements[level - 2]
    return gamatoto_xp


def edit_gamatoto_xp(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for gamatoto xp"""

    gamatoto_xp = save_stats["gamatoto_xp"]

    data = get_level_from_xp(gamatoto_xp["Value"], helper.check_data_is_jp(save_stats))
    level = data["level"]

    helper.colored_text(f"Gamatoto xp: &{gamatoto_xp['Value']}&\nLevel: &{level}&")
    raw = user_input_handler.colored_input(
        "Do you want to edit raw xp(&1&) or the level(&2&)?:"
    ) == "1"

    if raw:
        gam_xp = item.Item(
            name="Gamatoto XP",
            value=gamatoto_xp["Value"],
            max_value=None,
            edit_name="value",
        )
        gam_xp.edit()
        gamatoto_xp["Value"] = gam_xp.value
    else:
        gam_level = item.Item(
            name="Gamatoto Level",
            value=level,
            max_value=data["max_level"],
            edit_name="level",
        )
        gam_level.edit()
        gamatoto_xp["Value"] = get_xp_from_level(int(gam_level.value), helper.check_data_is_jp(save_stats))

    save_stats["gamatoto_xp"] = gamatoto_xp
    return save_stats
